from http import HTTPStatus


class CellxgeneException(Exception):
    """Base class for cellxgene exceptions"""

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class RequestException(CellxgeneException):
    """Baseclass for exceptions that can be raised from a request."""

    # The default status code is 400 (Bad Request)
    default_status_code = HTTPStatus.BAD_REQUEST

    def __init__(self, message, status_code=None):
        super().__init__(message)
        self.status_code = status_code or self.default_status_code


def define_exception(name, doc):
    globals()[name] = type(name, (CellxgeneException,), dict(__doc__=doc))


def define_request_exception(name, doc, default_status_code=HTTPStatus.BAD_REQUEST):
    globals()[name] = type(name, (RequestException,), dict(__doc__=doc, default_status_code=default_status_code))


define_request_exception("FilterError", "Raised when filter is malformed")
define_request_exception("JSONEncodingValueError", "Raised when data cannot be encoded into json")
define_request_exception("MimeTypeError", "Raised when incompatible MIME type selected")
define_request_exception("DatasetAccessError", "Raised when file loaded into a DataAdaptor is misformatted")
define_request_exception("DisabledFeatureError", "Raised when an attempt to use a disabled feature occurs")
define_request_exception("AnnotationsError", "Raised when an attempt to use the annotations feature fails")
define_request_exception(
    "ComputeError",
    "Raised when an error occurs during a compute algorithm (such as diffexp)",
    HTTPStatus.INTERNAL_SERVER_ERROR,
)
define_request_exception("ExceedsLimitError", "Raised when an HTTP request exceeds a limit/quota")
define_request_exception("ColorFormatException", "Raised when color helper functions encounter an unknown color format")
define_request_exception(
    "AuthenticationError", "Raised when there is an authentication error", default_status_code=HTTPStatus.UNAUTHORIZED
)

define_request_exception(
    "AnnotationCategoryNameError",
    "Raised when an annotation category name cant be saved",
    default_status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
)

define_exception("OntologyLoadFailure", "Raised when reading the ontology file fails")
define_exception("ConfigurationError", "Raised when checking configuration errors")
define_exception("PrepareError", "Raised when data is misprepared")
define_exception("SecretKeyRetrievalError", "Raised when get_secret_key from AWS fails")
define_exception("ObsoleteRequest", "Raised when the request is no longer valid.")
define_exception("UnsupportedSummaryMethod", "Raised when a gene set summary method is unknown or unsupported.")



import argparse
import collections
import json
import logging
import math
import string

import anndata
import numpy as np
import pandas as pd
import yaml

from . import gene_symbol
from . import ontology
from . import validate

REPLACE_SUFFIX = "_original"
ONTOLOGY_SUFFIX = "_ontology_term_id"


def is_curie(value):
    """Return True iff the value is an OBO-id CURIE like EFO:000001"""
    return (value.count(":")
            and all(len(part) > 0 for part in value.split(":"))
            and all(c in string.digits for c in value.split(":")[1]))


def is_ontology_field(field_name):
    """Return True iff the field_name is an ontology field like tissue_ontology_term_id"""
    return field_name.endswith(ONTOLOGY_SUFFIX)


def get_label_field_name(field_name):
    """Get the associated label field from an ontology field, assay_ontology_term_id --> assay"""
    return field_name[: -len(ONTOLOGY_SUFFIX)]


def split_suffix(maybe_curie):
    """Split off the (cell culture) or (organoid) suffix."""

    suffixes = [" (cell culture)", " (organoid)"]
    for suffix in suffixes:
        if maybe_curie.endswith(suffix):
            return maybe_curie[:-len(suffix)], suffix
    return maybe_curie, ""


def get_curie_and_label(maybe_curie):
    """Given a string that might be a curie, return a (curie, label) pair"""

    maybe_curie, suffix = split_suffix(maybe_curie)
    if not is_curie(maybe_curie):
        return ("", maybe_curie + suffix)
    return (maybe_curie + suffix, ontology.get_ontology_label(maybe_curie) + suffix)


def safe_add_field(adata_attr, field_name, field_value):
    """Add a field and value to an AnnData, but don't clobber an exising value."""

    if (
        isinstance(field_value, list)
        and field_value
        and isinstance(field_value[0], dict)
    ):
        field_value = json.dumps(field_value)
    if field_name in adata_attr:
        adata_attr[field_name + REPLACE_SUFFIX] = adata_attr[field_name]
    adata_attr[field_name] = field_value


def remix_uns(adata, uns_config):
    """Add fields from the config to adata.uns"""
    for field_name, field_value in uns_config.items():

        if is_ontology_field(field_name):
            # If it's an ontology field, look it up
            label_field_name = get_label_field_name(field_name)
            ontology_term, ontology_label = get_curie_and_label(field_value)
            safe_add_field(adata.uns, field_name, ontology_term)
            safe_add_field(adata.uns, label_field_name, ontology_label)
        else:
            safe_add_field(adata.uns, field_name, field_value)


def remix_obs(adata, obs_config):
    """Add fields from the config to adata.obs"""

    for field_name, field_value in obs_config.items():

        if isinstance(field_value, dict):
            # If the value is a dict, that means we are supposed to map from an
            # existing column to the new one
            source_column, column_map = next(iter(field_value.items()))
            nan_value = None
            for key in column_map:
                if isinstance(key, float) and math.isnan(key):
                    nan_value = column_map[key]
            if nan_value is not None:
                column_map["nan"] = nan_value

            for key in column_map:
                if key not in adata.obs[source_column].unique():
                    logging.warning(f'Key {key} not in adata.obs["{source_column}"]')

            for value in adata.obs[source_column].unique():
                if value not in column_map:
                    logging.warning(f'Value {value} in adata.obs["{source_column}"] not in translation dict')

            if is_ontology_field(field_name):
                ontology_term_map, ontology_label_map = {}, {}
                logging.info(f"Looking up labels for {field_name}")
                for original_value, maybe_curie in column_map.items():
                    curie, label = get_curie_and_label(maybe_curie)
                    ontology_term_map[original_value] = curie
                    ontology_label_map[original_value] = label
                    logging.info(f"Mapping {original_value} -> {curie} -> {label}")

                ontology_column = adata.obs[source_column].replace(
                    ontology_term_map, inplace=False
                )
                label_column = adata.obs[source_column].replace(
                    ontology_label_map, inplace=False
                )

                safe_add_field(adata.obs, field_name, ontology_column)
                safe_add_field(
                    adata.obs, get_label_field_name(field_name), label_column
                )
            else:
                label_column = adata.obs[source_column].replace(
                    column_map, inplace=False
                )
                safe_add_field(adata.obs, field_name, label_column)

        else:
            if is_ontology_field(field_name):
                # If it's an ontology field, look it up
                label_field_name = get_label_field_name(field_name)
                ontology_term, ontology_label = get_curie_and_label(field_value)
                safe_add_field(adata.obs, field_name, ontology_term)
                safe_add_field(adata.obs, label_field_name, ontology_label)
            else:
                safe_add_field(adata.obs, field_name, field_value)


def merge_df(df, domain, index, columns):
    """
    Given a dataframe with duplicate column labels, merge and return a dataframe where
    the duplicates have been merged together, resulting in a dataframe with unique column
    labels.

    "merge" depends on the value of domain. If the domain is "raw", then duplicate columns
    can just be summed. If it's "log1p" or "sqrt", it needs to be exp1m'd or squared, then
    summed, and then logged or sqrt'd again.
    """

    if not isinstance(df, np.ndarray):
        to_merge = df.toarray()
    else:
        to_merge = df
    if domain == "raw":
        merged_df = pd.DataFrame(to_merge, index=index, columns=columns).sum(
            axis=1, level=0, skipna=False
        )
    elif domain == "log1p":
        merged_df = (
            pd.DataFrame(np.expm1(to_merge, dtype=np.float128), index=index, columns=columns)
            .sum(axis=1, level=0, skipna=False)
        )
        merged_df = pd.DataFrame(np.log1p(merged_df.to_numpy()), index=merged_df.index, columns=merged_df.columns)
    elif domain == "sqrt":
        merged_df = (
            pd.DataFrame(np.square(to_merge), index=index, columns=columns)
            .sum(axis=1, level=0, skipna=False)
        )
        merged_df = pd.DataFrame(np.sqrt(merged_df.to_numpy()), index=merged_df.index, columns=merged_df.columns)

    return merged_df


def fixup_gene_symbols(adata, fixup_config):
    """Update the var index to hold a consistent set of HGNC gene symbols."""

    upgraded_var_index = gene_symbol.get_upgraded_var_index(adata.var)

    merged_X = merge_df(adata.X, fixup_config["X"], adata.obs.index, upgraded_var_index)
    fixup_adata = anndata.AnnData(
        X=merged_X,
        obs=adata.obs,
        var=merged_X.columns.to_frame(name="hgnc_gene_symbol"),
        uns=adata.uns,
        obsm=adata.obsm,
    )

    for layer, domain in fixup_config.items():
        if layer == "X":
            continue
        if layer == "raw.X":
            df = adata.raw.X
        else:
            df = adata.layers[layer]

        merged_df = merge_df(df, domain, adata.obs.index, upgraded_var_index)
        assert merged_df.index.equals(merged_X.index)
        assert merged_df.columns.equals(merged_X.columns)

        if domain == "raw":
            fixup_raw = anndata.AnnData(
                X=merged_df,
                obs=adata.obs,
                var=merged_X.columns.to_frame(name="hgnc_gene_symbol"),
            )
            fixup_adata.raw = fixup_raw
        else:
            fixup_adata.layers[layer] = merged_df

    return fixup_adata

def _strip_version(adata):
    """Remove version information from the AnnData object."""

    if "version" in adata.uns_keys():
        del adata.uns["version"]

def apply_schema(source_h5ad, remix_config, output_filename):

    try:
        import scanpy
    except ImportError:
        raise ImportError("scanpy must be installed for cellxgene schema")
    adata = scanpy.read_h5ad(source_h5ad)
    config = yaml.load(open(remix_config), Loader=yaml.FullLoader)
    remix_uns(adata, config["uns"])
    remix_obs(adata, config["obs"])

    if config.get("fixup_gene_symbols"):
        adata = fixup_gene_symbols(adata, config["fixup_gene_symbols"])

    if ("version" in adata.uns_keys()
            and isinstance(adata.uns["version"], collections.Mapping)
            and "corpora_schema_version" in adata.uns["version"]):
        schema_version = adata.uns["version"]["corpora_schema_version"]
        try:
            validate.get_schema_definition(schema_version)
        except ValueError:
            logging.warning(f"Stripping version information out of AnnData because schema "
                            f"version {schema_version} is unknown.")
            _strip_version(adata)

        if not validate.validate_adata(adata, shallow=False):
            logging.warning(f"Stripping version information out of AnnData because it does not "
                            f"follow schema version {schema_version} .")
            _strip_version(adata)

    adata.write_h5ad(output_filename, compression="gzip")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--source-h5ad", required=True)
    parser.add_argument("--remix-config", required=True)
    parser.add_argument("--output-filename", required=True)
    args = parser.parse_args()
    apply_schema(args.source_h5ad, args.remix_config, args.output_filename)



import os

from backend.server.common.config.base_config import BaseConfig
from backend.common.errors import ConfigurationError
from backend.server.common.config import get_secret_key
from backend.common.errors import SecretKeyRetrievalError
from backend.common.utils.type_conversion_utils import convert_string_to_value


class ExternalConfig(BaseConfig):
    """Manages the config attribute associated with external configuration sources, such as
    environment variables or the AWS Secrets Manager."""

    def __init__(self, app_config, default_config):
        super().__init__(app_config, default_config)
        try:
            self.environment = default_config["environment"]
            self.aws_secrets_manager__region = default_config["aws_secrets_manager"]["region"]
            self.aws_secrets_manager__secrets = default_config["aws_secrets_manager"]["secrets"]

        except KeyError as e:
            raise ConfigurationError(f"Unexpected config: {str(e)}")

    def complete_config(self, context):
        self.handle_environment(context)
        self.handle_aws_secrets_manager(context)

    def handle_environment(self, context):
        """For each environment variable defined, get the value (if it is set),
        and set the specified config parameter"""
        self.validate_correct_type_of_configuration_attribute("environment", list)
        for envdict in self.environment:
            name = envdict.get("name")
            if name is None:
                raise ConfigurationError("environment: 'name' is missing")
            required = envdict.get("required", False)
            if type(required) != bool:
                raise ConfigurationError("environment: 'required' must be a bool")
            path = envdict.get("path")
            if path is None:
                raise ConfigurationError("environment: 'path' is missing")

            value = os.environ.get(name)
            if value is None:
                if required:
                    raise ConfigurationError(f"required environment variable '{name}' not set")
            else:
                value = convert_string_to_value(value)
                self.app_config.update_single_config_from_path_and_value(path, value)

    def handle_aws_secrets_manager(self, context):
        """For each aws secret defined, get the key/values, and set the specified config parameter"""
        self.validate_correct_type_of_configuration_attribute("aws_secrets_manager__region", (type(None), str))
        self.validate_correct_type_of_configuration_attribute("aws_secrets_manager__secrets", list)

        if not self.aws_secrets_manager__secrets:
            return

        self.validate_correct_type_of_configuration_attribute("aws_secrets_manager__region", str)

        for secret in self.aws_secrets_manager__secrets:
            secret_name = secret.get("name")
            if secret_name is None:
                raise ConfigurationError("aws_secrets_manager: 'name' is missing")
            if not isinstance(secret_name, str):
                raise ConfigurationError("aws_secrets_manager: 'name' must be a string")

            try:
                secret_dict = get_secret_key(self.aws_secrets_manager__region, secret_name)
            except SecretKeyRetrievalError as e:
                raise ConfigurationError(f"Unable to retrieve secret {secret_name}: {str(e)}")

            values = secret.get("values")
            if values is None:
                raise ConfigurationError("aws_secrets_manager: 'values' is missing")
            if not isinstance(values, list):
                raise ConfigurationError("aws_secrets_manager: 'values' must be a list")

            for value in values:
                key = value.get("key")
                if key is None:
                    raise ConfigurationError(f"missing 'key' in secret values: {secret_name}")
                path = value.get("path")
                if path is None:
                    raise ConfigurationError(f"missing 'path' in secret values: {secret_name}")
                required = value.get("required", False)
                if type(required) != bool:
                    raise ConfigurationError(f"wrong type for 'required' in secret values: {secret_name}")

                secret_value = secret_dict.get(key)
                if secret_value is None:
                    if required:
                        raise ConfigurationError(f"required secret '{secret_name}:{key}' not set")
                else:
                    secret_value = convert_string_to_value(secret_value)
                    self.app_config.update_single_config_from_path_and_value(path, secret_value)



import logging

import boto3
from flask import json

from backend.common.errors import SecretKeyRetrievalError


def get_secret_key(region_name, secret_name):
    session = boto3.session.Session()
    client = session.client(service_name="secretsmanager", region_name=region_name)

    try:
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
        if "SecretString" in get_secret_value_response:
            var = get_secret_value_response["SecretString"]
            secret = json.loads(var)
            return secret
    except Exception as e:
        logging.critical(f"Caught exception during get_secret_key, {e}", exc_info=True)
        raise SecretKeyRetrievalError(str(e))

    return None



from backend.server.auth.auth import AuthTypeBase, AuthTypeFactory
from flask import session
from uuid import uuid4


class AuthTypeSession(AuthTypeBase):
    """Session based authentication.  The user is always logged.  The user id is a random number
    associated with the session.  This is a good choice for desktop servers."""

    # key in the session token for userid
    CXGUID = "cxguid"

    def __init__(self, app_config):
        super().__init__()

    def is_valid_authentication_type(self):
        return True

    def complete_setup(self, app):
        pass

    def is_user_authenticated(self):
        # always authenticated
        return True

    def get_user_id(self):
        if self.CXGUID not in session:
            session[self.CXGUID] = uuid4().hex
            session.permanent = True
        return session[self.CXGUID]
        
    def get_user_name(self):
        return "anonymous"

    def get_user_email(self):
        return None


AuthTypeFactory.register("session", AuthTypeSession)



import os
import sys
import warnings
from os.path import basename
from urllib.parse import urlparse

from backend.server.auth.auth import AuthTypeFactory
from backend.server.common.config.base_config import BaseConfig
from backend.server.common.config import DEFAULT_SERVER_PORT, BIG_FILE_SIZE_THRESHOLD
from backend.common.utils.data_locator import discover_s3_region_name
from backend.common.errors import ConfigurationError, DatasetAccessError
from backend.common.utils.utils import is_port_available, find_available_port, custom_format_warning
from backend.server.data_common.matrix_loader import MatrixDataLoader


class ServerConfig(BaseConfig):
    """Manages the config attribute associated with the server."""

    def __init__(self, app_config, default_config):
        super().__init__(app_config, default_config)

        try:
            self.app__verbose = default_config["app"]["verbose"]
            self.app__debug = default_config["app"]["debug"]
            self.app__host = default_config["app"]["host"]
            self.app__port = default_config["app"]["port"]
            self.app__open_browser = default_config["app"]["open_browser"]
            self.app__force_https = default_config["app"]["force_https"]
            self.app__flask_secret_key = default_config["app"]["flask_secret_key"]
            self.app__generate_cache_control_headers = default_config["app"]["generate_cache_control_headers"]

            self.authentication__type = default_config["authentication"]["type"]
            self.authentication__insecure_test_environment = default_config["authentication"][
                "insecure_test_environment"
            ]

            self.single_dataset__datapath = default_config["single_dataset"]["datapath"]
            self.single_dataset__obs_names = default_config["single_dataset"]["obs_names"]
            self.single_dataset__var_names = default_config["single_dataset"]["var_names"]
            self.single_dataset__about = default_config["single_dataset"]["about"]
            self.single_dataset__title = default_config["single_dataset"]["title"]

            self.data_locator__s3__region_name = default_config["data_locator"]["s3"]["region_name"]

            self.adaptor__anndata_adaptor__backed = default_config["adaptor"]["anndata_adaptor"]["backed"]

            self.limits__diffexp_cellcount_max = default_config["limits"]["diffexp_cellcount_max"]
            self.limits__column_request_max = default_config["limits"]["column_request_max"]

        except KeyError as e:
            raise ConfigurationError(f"Unexpected config: {str(e)}")

        self.data_adaptor = None

        # The authentication object
        self.auth = None

    def complete_config(self, context):
        self.handle_app(context)
        self.handle_data_source()
        self.handle_authentication()
        self.handle_data_locator()
        self.handle_adaptor()  # may depend on data_locator
        self.handle_single_dataset(context)  # may depend on adaptor
        self.handle_limits()

        self.check_config()

    def handle_app(self, context):
        self.validate_correct_type_of_configuration_attribute("app__verbose", bool)
        self.validate_correct_type_of_configuration_attribute("app__debug", bool)
        self.validate_correct_type_of_configuration_attribute("app__host", str)
        self.validate_correct_type_of_configuration_attribute("app__port", (type(None), int))
        self.validate_correct_type_of_configuration_attribute("app__open_browser", bool)
        self.validate_correct_type_of_configuration_attribute("app__force_https", bool)
        self.validate_correct_type_of_configuration_attribute("app__flask_secret_key", str)
        self.validate_correct_type_of_configuration_attribute("app__generate_cache_control_headers", bool)

        if self.app__port:
            try:
                if not is_port_available(self.app__host, self.app__port):
                    raise ConfigurationError(
                        f"The port selected {self.app__port} is in use, please configure an open port."
                    )
            except OverflowError:
                raise ConfigurationError(f"Invalid port: {self.app__port}")
        else:
            try:
                default_server_port = int(os.environ.get("CXG_SERVER_PORT", DEFAULT_SERVER_PORT))
            except ValueError:
                raise ConfigurationError(
                    "Invalid port from environment variable CXG_SERVER_PORT: " + os.environ.get("CXG_SERVER_PORT")
                )
            try:
                self.app__port = find_available_port(self.app__host, default_server_port)
            except OverflowError:
                raise ConfigurationError(f"Invalid port: {default_server_port}")

        if self.app__debug:
            context["messagefn"]("in debug mode, setting verbose=True and open_browser=False")
            self.app__verbose = True
            self.app__open_browser = False
        else:
            warnings.formatwarning = custom_format_warning

        if not self.app__verbose:
            sys.tracebacklimit = 0

    def handle_authentication(self):
        self.validate_correct_type_of_configuration_attribute("authentication__type", (type(None), str))
        self.validate_correct_type_of_configuration_attribute("authentication__insecure_test_environment", bool)

        if self.authentication__type == "test" and not self.authentication__insecure_test_environment:
            raise ConfigurationError("Test auth can only be used in an insecure test environment")

        self.auth = AuthTypeFactory.create(self.authentication__type, self)
        if self.auth is None:
            raise ConfigurationError(f"Unknown authentication type: {self.authentication__type}")

    def handle_data_locator(self):
        self.validate_correct_type_of_configuration_attribute("data_locator__s3__region_name", (type(None), bool, str))
        if self.data_locator__s3__region_name is True:
            path = self.single_dataset__datapath

            if path.startswith("s3://"):
                region_name = discover_s3_region_name(path)
                if region_name is None:
                    raise ConfigurationError(f"Unable to discover s3 region name from {path}")
            else:
                region_name = None
            self.data_locator__s3__region_name = region_name

    def handle_data_source(self):
        self.validate_correct_type_of_configuration_attribute("single_dataset__datapath", str)

    def handle_single_dataset(self, context):
        self.validate_correct_type_of_configuration_attribute("single_dataset__datapath", (str, type(None)))
        self.validate_correct_type_of_configuration_attribute("single_dataset__title", (str, type(None)))
        self.validate_correct_type_of_configuration_attribute("single_dataset__about", (str, type(None)))
        self.validate_correct_type_of_configuration_attribute("single_dataset__obs_names", (str, type(None)))
        self.validate_correct_type_of_configuration_attribute("single_dataset__var_names", (str, type(None)))

        # preload this data set
        matrix_data_loader = MatrixDataLoader(self.single_dataset__datapath, app_config=self.app_config)
        try:
            matrix_data_loader.pre_load_validation()
        except DatasetAccessError as e:
            raise ConfigurationError(str(e))

        file_size = matrix_data_loader.file_size()
        file_basename = basename(self.single_dataset__datapath)
        if file_size > BIG_FILE_SIZE_THRESHOLD:
            context["messagefn"](f"Loading data from {file_basename}, this may take a while...")
        else:
            context["messagefn"](f"Loading data from {file_basename}.")

        if self.single_dataset__about:

            def url_check(url):
                try:
                    result = urlparse(url)
                    if all([result.scheme, result.netloc]):
                        return True
                    else:
                        return False
                except ValueError:
                    return False

            if not url_check(self.single_dataset__about):
                raise ConfigurationError(
                    "Must provide an absolute URL for --about. (Example format: http://example.com)"
                )

    def handle_adaptor(self):
        self.validate_correct_type_of_configuration_attribute("adaptor__anndata_adaptor__backed", bool)

    def handle_limits(self):
        self.validate_correct_type_of_configuration_attribute("limits__diffexp_cellcount_max", (type(None), int))
        self.validate_correct_type_of_configuration_attribute("limits__column_request_max", (type(None), int))

    def exceeds_limit(self, limit_name, value):
        limit_value = getattr(self, "limits__" + limit_name, None)
        if limit_value is None:  # disabled
            return False
        return value > limit_value



from os.path import expanduser, isdir, isfile, sep, splitext

import click
import pandas as pd
from numpy import ndarray, unique
from scipy.sparse.csc import csc_matrix

from backend.common.utils.utils import sort_options


@sort_options
@click.command(
    short_help="Preprocess data for use with cellxgene. " "Run `cellxgene prepare --help` for more information.",
    options_metavar="<options>",
)
@click.argument("data", nargs=1, metavar="<path to data file>", required=True)
@click.option(
    "--embedding",
    "-e",
    default=["umap", "tsne"],
    multiple=True,
    type=click.Choice(["umap", "tsne"]),
    help="Embedding algorithm(s). Repeat option for multiple embeddings.",
    show_default=True,
)
@click.option(
    "--recipe", "-r", default="none", type=click.Choice(["none", "seurat", "zheng17"]), show_default=True,
)
@click.option("--output", "-o", default="", help="Save a new file to filename.", metavar="<filename>")
@click.option("--plotting", "-p", default=False, is_flag=True, help="Generate plots.", show_default=True)
@click.option("--sparse", default=False, is_flag=True, help="Force sparsity.", show_default=True)
@click.option("--overwrite", default=False, is_flag=True, help="Allow file overwriting.", show_default=True)
@click.option("--set-obs-names", default="", help="Named field to set as index for obs.", metavar="<name>")
@click.option("--set-var-names", default="", help="Named field to set as index for var.", metavar="<name>")
@click.option(
    "--skip-qc",
    default=False,
    is_flag=True,
    help="Do not run quality control metrics. By default cellxgene runs them "
    "(saved to adata.obs and adata.var; see scanpy.pp.calculate_qc_metrics for details).",
)
@click.option(
    "--make-obs-names-unique/--no-make-obs-names-unique",
    default=True,
    help="Ensure obs index is unique.",
    show_default=True,
)
@click.option(
    "--make-var-names-unique/--no-make-var-names-unique",
    default=True,
    help="Ensure var index is unique.",
    show_default=True,
)
@click.help_option("--help", "-h", help="Show this message and exit.")
def prepare(
    data,
    embedding,
    recipe,
    output,
    plotting,
    sparse,
    overwrite,
    set_obs_names,
    set_var_names,
    skip_qc,
    make_obs_names_unique,
    make_var_names_unique,
):
    """
    Preprocess data for use with cellxgene.
    This tool runs a series of scanpy routines for preparing a dataset for use
    with cellxgene. It loads data from different formats
    (h5ad, loom, or a 10x directory), runs dimensionality reduction,
    computes nearest neighbors, computes an embedding, performs clustering,
    and saves the results. Includes additional options for naming annotations,
    ensuring sparsity, and plotting results.
    """

    # collect slow imports here to make CLI startup more responsive
    click.echo("[cellxgene] Starting CLI...")
    try:
        import matplotlib

        matplotlib.use("Agg")
        import scanpy as sc
    except ImportError:
        raise click.ClickException(
            "[cellxgene] cellxgene prepare has not been installed. Please run `pip install 'cellxgene[prepare]'` "
            "to install the necessary requirements."
        )

    # scanpy settings
    sc.settings.verbosity = 0
    sc.settings.autosave = True

    # check args
    if sparse and not recipe == "none":
        raise click.UsageError("Cannot use a recipe when forcing sparsity")

    output = expanduser(output)

    if not output:
        click.echo(
            "Warning: No file will be saved, to save the results of cellxgene prepare include "
            "--output <filename> to save output to a new file"
        )
    if isfile(output) and not overwrite:
        raise click.UsageError(f"Cannot overwrite existing file {output}, try using the flag --overwrite")

    def load_data(data):
        if isfile(data):
            name, extension = splitext(data)
            if extension == ".h5ad":
                adata = sc.read_h5ad(data)
            elif extension == ".loom":
                adata = sc.read_loom(data)
            else:
                raise click.FileError(data, hint="does not have a valid extension [.h5ad | .loom]")
        elif isdir(data):
            if not data.endswith(sep):
                data += sep
            adata = sc.read_10x_mtx(data)
        else:
            raise click.FileError(data, hint="not a valid file or path")

        if not set_obs_names == "":
            if set_obs_names not in adata.obs_keys():
                raise click.UsageError(f"obs {set_obs_names} not found, options are: {adata.obs_keys()}")
            adata.obs_names = adata.obs[set_obs_names]
        if not set_var_names == "":
            if set_var_names not in adata.var_keys():
                raise click.UsageError(f"var {set_var_names} not found, options are: {adata.var_keys()}")
            adata.var_names = adata.var[set_var_names]
        if make_obs_names_unique:
            adata.obs.index = make_index_unique(adata.obs.index)
        if make_var_names_unique:
            adata.var.index = make_index_unique(adata.var.index)
        if not adata._obs.index.is_unique:
            click.echo("Warning: obs index is not unique")
        if not adata._var.index.is_unique:
            click.echo("Warning: var index is not unique")
        return adata

    def calculate_qc_metrics(adata):
        if not skip_qc:
            sc.pp.calculate_qc_metrics(adata, inplace=True)
        return adata

    def make_sparse(adata):
        if (type(adata.X) is ndarray) and sparse:
            adata.X = csc_matrix(adata.X)

    def run_recipe(adata):
        if recipe == "seurat":
            sc.pp.recipe_seurat(adata)
        elif recipe == "zheng17":
            sc.pp.recipe_zheng17(adata)
        else:
            sc.pp.filter_cells(adata, min_genes=5)
            sc.pp.filter_genes(adata, min_cells=25)
            if sparse:
                sc.pp.scale(adata, zero_center=False)
            else:
                sc.pp.scale(adata)

    def run_pca(adata):
        if sparse:
            sc.pp.pca(adata, svd_solver="arpack", zero_center=False)
        else:
            sc.pp.pca(adata, svd_solver="arpack")

    def run_neighbors(adata):
        sc.pp.neighbors(adata)

    def run_louvain(adata):
        sc.tl.louvain(adata)

    def run_embedding(adata):
        if len(unique(adata.obs["louvain"].values)) < 10:
            palette = "tab10"
        else:
            palette = "tab20"

        if "umap" in embedding:
            sc.tl.umap(adata)
            if plotting:
                sc.pl.umap(adata, color="louvain", palette=palette, save="_louvain")

        if "tsne" in embedding:
            sc.tl.tsne(adata)
            if plotting:
                sc.pl.tsne(adata, color="louvain", palette=palette, save="_louvain")

    def show_step(item):
        if not skip_qc:
            qc_name = "Calculating QC metrics"
        else:
            qc_name = "Skipping QC"
        names = {
            "calculate_qc_metrics": qc_name,
            "make_sparse": "Ensuring sparsity",
            "run_recipe": f'Running preprocessing recipe "{recipe}"',
            "run_pca": "Running PCA",
            "run_neighbors": "Calculating neighbors",
            "run_louvain": "Calculating clusters",
            "run_embedding": "Computing embedding",
        }
        if item is not None:
            return names[item.__name__]

    steps = [calculate_qc_metrics, make_sparse, run_recipe, run_pca, run_neighbors, run_louvain, run_embedding]

    click.echo(f"[cellxgene] Loading data from {data}, please wait...")
    adata = load_data(data)

    click.echo("[cellxgene] Beginning preprocessing...")
    with click.progressbar(steps, label="[cellxgene] Progress", show_eta=False, item_show_func=show_step) as bar:
        for step in bar:
            step(adata)

    # saving
    if not output == "":
        click.echo(f"[cellxgene] Saving results to {output}...")
        adata.write(output)

    click.echo("[cellxgene] Success!")


# TODO (mweiden): remove this once this issue is resolved https://github.com/theislab/anndata/issues/344
# Note: tentative solution here https://github.com/theislab/anndata/pull/345
def make_index_unique(index: pd.Index, join: str = "-"):
    """
    Makes the index unique by appending a number string to each duplicate index element: '1', '2', etc.

    If a tentative name created by the algorithm already exists in the index, it tries the next integer in the sequence.

    The first occurrence of a non-unique value is ignored.
    Parameters
    ----------
    join
         The connecting string between name and integer.
    Examples
    --------
    >>> from anndata import AnnData
    >>> adata1 = AnnData(np.ones((3, 2)), dict(obs_names=['a', 'b', 'c']))
    >>> adata2 = AnnData(np.zeros((3, 2)), dict(obs_names=['d', 'b', 'b']))
    >>> adata = adata1.concatenate(adata2)
    >>> adata.obs_names
    Index(['a', 'b', 'c', 'd', 'b', 'b'], dtype='object')
    >>> adata.obs_names_make_unique()
    >>> adata.obs_names
    Index(['a', 'b', 'c', 'd', 'b-1', 'b-2'], dtype='object')
    """
    if index.is_unique:
        return index
    from collections import defaultdict

    values = index.values
    values_set = set(values)
    indices_dup = index.duplicated(keep="first")
    values_dup = values[indices_dup]
    counter = defaultdict(lambda: 0)
    for i, v in enumerate(values_dup):
        while True:
            counter[v] += 1
            tentative_new_name = v + join + str(counter[v])
            if tentative_new_name not in values_set:
                values_set.add(tentative_new_name)
                values_dup[i] = tentative_new_name
                break

    values[indices_dup] = values_dup
    index = pd.Index(values)
    return index



from abc import ABCMeta, abstractmethod
from os.path import basename, splitext
import numpy as np
import pandas as pd
from scipy import sparse
from server_timing import Timing as ServerTiming

from backend.server.common.config.app_config import AppConfig
from backend.common.constants import Axis
from backend.common.errors import FilterError, JSONEncodingValueError, ExceedsLimitError, UnsupportedSummaryMethod
from backend.common.utils.utils import jsonify_numpy
from backend.common.fbs.matrix import encode_matrix_fbs
from backend.common.genesets import validate_gene_sets


class DataAdaptor(metaclass=ABCMeta):
    """Base class for loading and accessing matrix data"""

    def __init__(self, data_locator, app_config, dataset_config=None):
        if not isinstance(app_config, AppConfig):
            raise TypeError("config expected to be of type AppConfig")

        # location to the dataset
        self.data_locator = data_locator

        # config is the application configuration
        self.app_config = app_config
        self.server_config = self.app_config.server_config
        self.dataset_config = dataset_config or app_config.dataset_config

        # parameters set by this data adaptor based on the data.
        self.parameters = {}

    @staticmethod
    @abstractmethod
    def pre_load_validation(data_locator):
        pass

    @staticmethod
    @abstractmethod
    def open(data_locator, app_config, dataset_config):
        pass

    @staticmethod
    @abstractmethod
    def file_size(data_locator):
        pass

    @abstractmethod
    def get_name(self):
        """return a string name for this data adaptor"""
        pass

    @abstractmethod
    def get_library_versions(self):
        """return a dictionary of library name to library versions"""
        pass

    @abstractmethod
    def get_embedding_names(self):
        """return a list of pre-computed embedding names"""
        pass

    @abstractmethod
    def get_embedding_array(self, ename, dims=2):
        """return an numpy array for the given pre-computed embedding name."""
        pass

    
    @abstractmethod
    def compute_sankey_df(self, labels, name):
        """compute sankey"""
        pass

    @abstractmethod
    def compute_embedding(self, method, filter, reembedParams):
        """compute a new embedding on the specified obs subset and return the embedding schema."""
        pass

    @abstractmethod
    def get_X_array(self, col_idx):
        """return the X array, possibly filtered by obs_mask or var_mask.
        the return type is either ndarray or scipy.sparse.spmatrix."""
        pass

    @abstractmethod
    def get_shape(self):
        pass

    @abstractmethod
    def query_var_array(self, term_var):
        pass

    @abstractmethod
    def query_obs_array(self, term_var):
        pass

    @abstractmethod
    def get_colors(self):
        pass

    @abstractmethod
    def get_obs_index(self):
        pass

    @abstractmethod
    def get_obs_columns(self):
        pass

    @abstractmethod
    def get_obs_keys(self):
        # return list of keys
        pass

    @abstractmethod
    def get_var_keys(self):
        # return list of keys
        pass

    @abstractmethod
    def cleanup(self):
        pass

    def get_data_locator(self):
        return self.data_locator

    def get_location(self):
        return self.data_locator.uri_or_path

    def get_about(self):
        return None

    def get_title(self):
        # default to file name
        location = self.get_location()
        if location.endswith("/"):
            location = location[:-1]
        return splitext(basename(location))[0]

    def get_corpora_props(self):
        return None

    @abstractmethod
    def get_schema(self):
        """
        Return current schema
        """
        pass

    @abstractmethod
    def annotation_to_fbs_matrix(self, axis, field=None, uid=None):
        """
        Gets annotation value for each observation
        :param axis: string obs or var
        :param fields: list of keys for annotation to return, returns all annotation values if not set.
        :return: flatbuffer: in fbs/matrix.fbs encoding
        """
        pass

    def update_parameters(self, parameters):
        parameters.update(self.parameters)

    def _index_filter_to_mask(self, filter, count):
        mask = np.zeros((count,), dtype=np.bool)
        for i in filter:
            if isinstance(i, list):
                mask[i[0] : i[1]] = True
            else:
                mask[i] = True
        return mask

    def _axis_filter_to_mask(self, axis, filter, count):
        mask = np.ones((count,), dtype=np.bool)
        if "index" in filter:
            mask = np.logical_and(mask, self._index_filter_to_mask(filter["index"], count))
        if "annotation_value" in filter:
            mask = np.logical_and(mask, self._annotation_filter_to_mask(axis, filter["annotation_value"], count))

        return mask

    def _annotation_filter_to_mask(self, axis, filter, count):
        mask = np.ones((count,), dtype=np.bool)
        for v in filter:
            name = v["name"]
            if axis == Axis.VAR:
                anno_data = self.query_var_array(name)
            elif axis == Axis.OBS:
                anno_data = self.query_obs_array(name)

            if anno_data.dtype.name in ["boolean", "category", "object"]:
                values = v.get("values", [])
                key_idx = np.in1d(anno_data, values)
                mask = np.logical_and(mask, key_idx)

            else:
                min_ = v.get("min", None)
                max_ = v.get("max", None)
                if min_ is not None:
                    key_idx = (anno_data >= min_).ravel()
                    mask = np.logical_and(mask, key_idx)
                if max_ is not None:
                    key_idx = (anno_data <= max_).ravel()
                    mask = np.logical_and(mask, key_idx)

        return mask

    def _filter_to_mask(self, filter):
        """
        Return the filter as a row and column selection list.
        No filter on a dimension means 'all'
        """
        shape = self.get_shape()
        var_selector = None
        obs_selector = None
        if filter is not None:
            if Axis.OBS in filter:
                obs_selector = self._axis_filter_to_mask(Axis.OBS, filter["obs"], shape[0])

            if Axis.VAR in filter:
                var_selector = self._axis_filter_to_mask(Axis.VAR, filter["var"], shape[1])

        return (obs_selector, var_selector)

    def check_new_labels(self, labels_df):
        """Check the new annotations labels, then set the labels_df index"""
        if labels_df is None or labels_df.empty:
            return

        labels_df.index = self.get_obs_index()
        if labels_df.index.name is None:
            labels_df.index.name = "index"

        # all labels must have a name, which must be unique and not used in obs column names
        if not labels_df.columns.is_unique:
            raise KeyError("All column names specified in user annotations must be unique.")

        # the label index must be unique, and must have same values the anndata obs index
        if not labels_df.index.is_unique:
            raise KeyError("All row index values specified in user annotations must be unique.")

        # labels must have same count as obs annotations
        shape = self.get_shape()
        if labels_df.shape[0] != shape[0]:
            raise ValueError("Labels file must have same number of rows as data file.")

        # This will convert a float column that contains integer data into an integer type.
        # This case can occur when a user makes a copy of a category that originally contained integer data.
        # The client always copies array data to floats, therefore the copy will contain floats instead of integers.
        # float data is not allowed as a categorical type.
        if any([np.issubdtype(coltype.type, np.floating) for coltype in labels_df.dtypes]):
            labels_df = labels_df.convert_dtypes()
            for col, dtype in zip(labels_df, labels_df.dtypes):
                if isinstance(dtype, pd.Int32Dtype):
                    labels_df[col] = labels_df[col].astype("int32")
                if isinstance(dtype, pd.Int64Dtype):
                    labels_df[col] = labels_df[col].astype("int64")
        
        return labels_df

    def check_new_gene_sets(self, genesets, context=None):
        var_names = set(self.query_var_array(self.parameters.get("var_names")))
        return validate_gene_sets(genesets, var_names)

    def data_frame_to_fbs_matrix(self, filter, axis, layer="X", logscale=False):
        """
        Retrieves data 'X' and returns in a flatbuffer Matrix.
        :param filter: filter: dictionary with filter params
        :param axis: string obs or var
        :return: flatbuffer Matrix

        Caveats:
        * currently only supports access on VAR axis
        * currently only supports filtering on VAR axis
        """
        if axis != Axis.VAR:
            raise ValueError("Only VAR dimension access is supported")
        try:
            d = filter['var']['annotation_value'][0]
            col = d['name']
            vals = d['values']
            var_selector = np.in1d(np.array(list(self.data.var[col])),vals)
        except (KeyError, IndexError, TypeError, AttributeError):
            raise FilterError("Error parsing filter")

        col_idx = np.nonzero([] if var_selector is None else var_selector)[0]
        X = self.get_X_array(col_idx,layer=layer,logscale=logscale)
        return encode_matrix_fbs(X, col_idx=col_idx, row_idx=None)

    def diffexp_topN(self, obsFilterA, obsFilterB, top_n=None):
        """
        Computes the top N differentially expressed variables between two observation sets. If mode
        is "TOP_N", then stats for the top N
        dataframes
        contain a subset of variables, then statistics for all variables will be returned, otherwise
        only the top N vars will be returned.
        :param obsFilterA: filter: dictionary with filter params for first set of observations
        :param obsFilterB: filter: dictionary with filter params for second set of observations
        :param top_n: Limit results to top N (Top var mode only)
        :return: top N genes and corresponding stats
        """
        if Axis.VAR in obsFilterA or Axis.VAR in obsFilterB:
            raise FilterError("Observation filters may not contain variable conditions")
        try:
            shape = self.get_shape()
            obs_mask_A = self._axis_filter_to_mask(Axis.OBS, obsFilterA["obs"], shape[0])
            obs_mask_B = self._axis_filter_to_mask(Axis.OBS, obsFilterB["obs"], shape[0])
        except (KeyError, IndexError):
            raise FilterError("Error parsing filter")
        if top_n is None:
            top_n = self.dataset_config.diffexp__top_n

        if self.server_config.exceeds_limit(
            "diffexp_cellcount_max", np.count_nonzero(obs_mask_A) + np.count_nonzero(obs_mask_B)
        ):
            raise ExceedsLimitError("Diffexp request exceeds max cell count limit")

        result = self.compute_diffexp_ttest(
            maskA=obs_mask_A,
            maskB=obs_mask_B,
            top_n=top_n,
            lfc_cutoff=self.dataset_config.diffexp__lfc_cutoff,
        )

        try:
            return jsonify_numpy(result)
        except ValueError:
            raise JSONEncodingValueError("Error encoding differential expression to JSON")

    @abstractmethod
    def compute_diffexp_ttest(self, maskA, maskB, top_n, lfc_cutoff):
        pass

    @staticmethod
    def normalize_embedding(embedding):
        """Normalize embedding layout to meet client assumptions.
        Embedding is an ndarray, shape (n_obs, n)., where n is normally 2
        """

        # scale isotropically
        try:
            min = np.nanmin(embedding, axis=0)
            max = np.nanmax(embedding, axis=0)
        except RuntimeError:
            # indicates entire array was NaN, which should propagate
            min = np.NaN
            max = np.NaN

        diff = max-min
        if np.abs(diff).sum()==0:
            normalized_layout = embedding
            normalized_layout[np.invert(np.isnan(normalized_layout))] = 0.5
        else:
            scale = np.amax(max - min)
            normalized_layout = (embedding - min) / scale

            # translate to center on both axis
            translate = 0.5 - ((max - min) / scale / 2)
            normalized_layout = normalized_layout + translate

            normalized_layout = normalized_layout.astype(dtype=np.float32)
        return normalized_layout

    def layout_to_fbs_matrix(self, fields):
        """
        return specified embeddings as a flatbuffer, using the cellxgene matrix fbs encoding.

        * returns only first two dimensions, with name {ename}_0 and {ename}_1,
          where {ename} is the embedding name.
        * client assumes each will be individually centered & scaled (isotropically)
          to a [0, 1] range.
        * does not support filtering

        """
        embeddings = self.get_embedding_names() if fields is None or len(fields) == 0 else fields
        layout_data = []
        with ServerTiming.time("layout.query"):
            for ename in embeddings:
                embedding = self.get_embedding_array(ename, 2)
                normalized_layout = DataAdaptor.normalize_embedding(embedding)
                layout_data.append(pd.DataFrame(normalized_layout, columns=[f"{ename}_0", f"{ename}_1"]))

        with ServerTiming.time("layout.encode"):
            if layout_data:
                df = pd.concat(layout_data, axis=1, copy=False)
            else:
                df = pd.DataFrame()
            fbs = encode_matrix_fbs(df, col_idx=df.columns, row_idx=None)

        return fbs

    def get_last_mod_time(self):
        try:
            lastmod = self.get_data_locator().lastmodtime()
        except RuntimeError:
            lastmod = None
        return lastmod

    def summarize_var(self, method, filter, query_hash,layer="X",logscale=False):
        if method != "mean":
            raise UnsupportedSummaryMethod("Unknown gene set summary method.")

        d = filter['var']['annotation_value'][0]
        col = d['name']
        vals = d['values']
        var_selector = np.in1d(np.array(list(self.data.var[col])),vals)
        if var_selector is None or np.count_nonzero(var_selector) == 0:
            mean = np.zeros((self.get_shape()[0], 1), dtype=np.float32)
        else:
            col_idx = np.nonzero([] if var_selector is None else var_selector)[0]
            X = self.get_X_array(col_idx,layer=layer,logscale=logscale)
            if sparse.issparse(X):
                mean = X.mean(axis=1)
            else:
                mean = X.mean(axis=1, keepdims=True)

        col_idx = pd.Index([query_hash])
        return encode_matrix_fbs(mean, col_idx=col_idx, row_idx=None)



import click

from backend.server.converters.schema import remix, validate


@click.group(
    name="schema",
    subcommand_metavar="COMMAND <args>",
    short_help="Apply and validate the cellxgene data integration schema to an h5ad file.",
    context_settings=dict(max_content_width=85, help_option_names=["-h", "--help"]),
)
def schema_cli():
    try:
        import scanpy  # noqa: F401
    except ImportError:
        raise click.ClickException(
            "[cellxgene] cellxgene schema requires scanpy"
        )


@click.command(
    name="apply",
    short_help="(experimental) Apply the cellxgene data integration schema to an h5ad.",
    help="(experimental) Using a yaml file that describes schema values to insert or convert and in input "
    "h5ad file, apply the schema changes and create a new, conforming h5ad.",
)
@click.option(
    "--source-h5ad",
    help="Input h5ad file.",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--remix-config",
    help="Config yaml with information on how to apply the schema.",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--output-filename",
    help="Filename for the new, schema-conforming h5ad file.",
    required=True,
    nargs=1
)
def schema_apply(source_h5ad, remix_config, output_filename):
    remix.apply_schema(source_h5ad, remix_config, output_filename)


@click.command(
    name="validate",
    short_help="(experimental) Check that an h5ad follows the cellxgene data integration schema.",
)
@click.argument(
    "h5ad",
    nargs=1,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--shallow",
    help="When true, just check that the correct version information is present.",
    default=False,
    show_default=True,
    is_flag=True,
)
def schema_validate(h5ad, shallow):
    validate.validate(h5ad, shallow)


schema_cli.add_command(schema_apply)
schema_cli.add_command(schema_validate)



import logging

import numpy as np
import pandas as pd


def get_dtypes_and_schemas_of_dataframe(dataframe: pd.DataFrame):
    dtypes_by_column_name = {}
    schema_type_hints_by_column_name = {}

    for column_name, column_values in dataframe.items():
        (
            dtypes_by_column_name[column_name],
            schema_type_hints_by_column_name[column_name],
        ) = get_dtype_and_schema_of_array(column_values)

    return dtypes_by_column_name, schema_type_hints_by_column_name


def get_dtype_of_array(array: pd.Series):
    return get_dtype_and_schema_of_array(array)[0]


def get_schema_type_hint_of_array(array: pd.Series):
    return get_dtype_and_schema_of_array(array)[1]


def get_dtype_and_schema_of_array(array: pd.Series):
    return (
        get_dtype_from_dtype(array.dtype, array_values=array),
        get_schema_type_hint_from_dtype(array.dtype, array_values=array),
    )


def get_dtype_from_dtype(dtype, array_values=None):
    """
    Given a data type, finds the equivalent data type that the array should be encoded as. Notably, this is relevant
    for 64 bit values which will get downcast to 32 bit.
    """

    dtype_name = dtype.name
    dtype_kind = dtype.kind

    if dtype_name == "bool":
        return np.uint8
    if dtype_name == "object" and dtype_kind == "O":
        return str
    if dtype_name == "category":
        return get_dtype_from_dtype(dtype.categories.dtype, array_values)

    if can_cast_to_int32(dtype, array_values):
        return np.int32
    if can_cast_to_float32(dtype, array_values):
        return np.float32
    if not can_cast_to_float32(dtype, array_values):
        return np.float64

    raise TypeError(f"Annotations of type {dtype} are unsupported.")


def get_schema_type_hint_from_dtype(dtype, array_values=None):
    """
    Returns a dictionary that contains type hints about the data type given, especially if the data type is 64 bit
    and will be downcast to 32 bit.
    """

    dtype_name = dtype.name
    dtype_kind = dtype.kind

    if dtype == np.float32 or dtype == np.int32:
        return {"type": dtype_name}
    if dtype_name == "bool":
        return {"type": "boolean"}
    if dtype_name == "object" and dtype_kind == "O":
        return {"type": "string"}
    if dtype_name == "category":
        return {"type": "categorical", "categories": dtype.categories.tolist()}

    if can_cast_to_int32(dtype, array_values):
        return {"type": "int32"}
    if can_cast_to_float32(dtype, array_values):
        return {"type": "float32"}
    if dtype_kind == "f" and not can_cast_to_float32(dtype, array_values):
        return {"type": "float64"}

    raise TypeError(f"Annotations of type {dtype} are unsupported.")


def can_cast_to_float32(dtype, array_values):
    """
    Optimistically returns True signifying that a type downcast to float32 is possible whenever the incoming type is
    a float.

    We also handle a special case here where the array is a Series object with integer categorical values AND NaNs.
    Since NaNs are floating points in numpy, we upcast the integer array to float32 and return True.
    """

    if dtype.kind == "f":
        if not np.can_cast(dtype, np.float32):
            logging.warning(f"Type {dtype.name} will be converted to 32 bit float and may lose precision.")

        return True

    if dtype.kind == "O" and array_values.hasnans:
        return True

    return False


def can_cast_to_int32(dtype, array_values=None):
    """
    A type can be cast to 32 bit, overriding the numpy `cast_cast` function if the values in the array that are of
    the higher precision type has values that are entirely within the range of the downcast type.
    """

    # Since a NaN is technically a float, any array that contains NaNs cannot be cast to an integer so immediately
    # return False.
    if array_values.hasnans:
        return False

    # If the array is categorical, then we need to order the array values so that functions min and max that occur
    # later, can function. They do not function on unordered categories.
    ordered_array_values = array_values
    if array_values.dtype.name == "category" and not array_values.cat.ordered:
        ordered_array_values = array_values.cat.as_ordered()

    if dtype.kind in ["i", "u"]:
        if np.can_cast(dtype, np.int32):
            return True
        ii32 = np.iinfo(np.int32)
        if (
            not ordered_array_values.empty
            and (ordered_array_values.min() >= ii32.min and ordered_array_values.max() <= ii32.max)
            or ordered_array_values.empty
        ):
            return True
    return False


def convert_pandas_series_to_numpy(series_to_convert: pd.Series, dtype):
    if series_to_convert.hasnans and dtype == np.int32:
        logging.error("Cannot convert a pandas Series object to an integer dtype if it contains NaNs.")

    return series_to_convert.to_numpy(dtype)


def convert_string_to_value(value: str):
    """convert a string to value with the most appropriate type"""
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    if value == "null":
        return None
    try:
        return eval(value)
    except:  # noqa E722
        return value



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: NetEncoding

class TypedArray(object):
    NONE = 0
    Float32Array = 1
    Int32Array = 2
    Uint32Array = 3
    Float64Array = 4
    JSONEncodedArray = 5




from backend.server.auth.auth import AuthTypeClientBase, AuthTypeFactory
from flask import session, request, redirect


class AuthTypeTest(AuthTypeClientBase):
    """An authentication type for testing client based logins.  When the login route is accessed
    the user is automatically logged in with a default or configured username"""

    # key in session token with userid and username
    CXGUID = "cxguid_test"
    CXGUNAME = "cxguname_test"
    CXGUEMAIL = "cxguemail_test"
    CXGUPICTURE = "cxgupicture_test"

    def __init__(self, app_config):
        super().__init__()
        self.user_name = "test_account"
        self.user_id = "id0001"
        self.user_email = "test_account@test.com"
        self.user_picture = None

    def is_valid_authentication_type(self):
        return True

    def requires_client_login(self):
        return True

    def add_url_rules(self, app):
        app.add_url_rule("/login", "login", self.login, methods=["GET"])
        app.add_url_rule("/logout", "logout", self.logout, methods=["GET"])

    def complete_setup(self, app):
        pass

    def is_user_authenticated(self):
        return self.CXGUID in session

    def get_user_id(self):
        return session.get(self.CXGUID)

    def get_user_name(self):
        return session.get(self.CXGUNAME)

    def get_user_email(self):
        return session.get(self.CXGUEMAIL)

    def get_user_picture(self):
        return session.get(self.CXGUPICTURE)

    def login(self):
        args = request.args
        return_to = args.get("dataset", "/")
        session[self.CXGUID] = args.get("userid", self.user_id)
        session[self.CXGUNAME] = args.get("username", self.user_name)
        session[self.CXGUEMAIL] = args.get("email", self.user_email)
        session[self.CXGUPICTURE] = args.get("picture", self.user_picture)
        return redirect(return_to)

    def logout(self):
        session.clear()
        return_to = request.args.get("dataset", "/")
        return redirect(return_to)

    def get_login_url(self, data_adaptor):
        """Return the url for the login route"""
        return "/login"

    def get_logout_url(self, data_adaptor):
        """Return the url for the logout route"""
        return "/logout"


AuthTypeFactory.register("test", AuthTypeTest)



import importlib
import numpy as np
from anndata import AnnData
"""
Wrapper for various scanpy modules.  Will raise NotImplementedError if the scanpy
module is not installed/available
"""


def get_scanpy_module():
    try:
        sc = importlib.import_module("scanpy")
        # Future: we could enforce versions here, eg, lookat sc.__version__
        return sc
    except ModuleNotFoundError as e:
        raise NotImplementedError("Please install scanpy to enable re-embedding.") from e
    except Exception as e:
        # will capture other ImportError corner cases
        raise NotImplementedError() from e

def get_scanpy_external_module():
    try:
        sc = importlib.import_module("scanpy")
        # Future: we could enforce versions here, eg, lookat sc.__version__
        return sc.external
    except ModuleNotFoundError as e:
        raise NotImplementedError("Please install scanpy to enable re-embedding.") from e
    except Exception as e:
        # will capture other ImportError corner cases
        raise NotImplementedError() from e

def get_samalg_module():
    try:
        samalg = importlib.import_module("samalg")
        # Future: we could enforce versions here, eg, lookat sc.__version__
        return samalg.SAM
    except ModuleNotFoundError as e:
        raise NotImplementedError("Please install sam-alogirthm to enable SAM during re-embedding") from e
    except Exception as e:
        # will capture other ImportError corner cases
        raise NotImplementedError() from e

def scanpy_umap(adata, obs_mask=None, reembedParams = {}, pca_options={}, neighbors_options={}, umap_options={}):
    """
    Given adata and an obs mask, return a new embedding for adata[obs_mask, :]
    as an ndarray of shape (len(obs_mask), N), where N>=2.

    Do NOT mutate adata.
    """

    # backed mode is incompatible with the current implementation
    if adata.isbacked:
        raise NotImplementedError("Backed mode is incompatible with re-embedding")

    # safely get scanpy module, which may not be present.
    sc = get_scanpy_module()

    # https://github.com/theislab/anndata/issues/311
    obs_mask = slice(None) if obs_mask is None else obs_mask
    adata = adata[obs_mask, :].copy()

    for k in list(adata.obsm.keys()):
        del adata.obsm[k]
    for k in list(adata.uns.keys()):
        del adata.uns[k]
    
    doPreprocess = reembedParams.get("doPreprocess",False)
    minCountsCF = reembedParams.get("minCountsCF",0)
    minGenesCF = reembedParams.get("minGenesCF",0)
    minCellsGF = reembedParams.get("minCellsGF",0)
    maxCellsGF = reembedParams.get("maxCellsGF",100)
    minCountsGF = reembedParams.get("minCountsGF",0)
    doSAM = reembedParams.get("doSAM",False)
    nTopGenesHVG = reembedParams.get("nTopGenesHVG",2000)
    nBinsHVG = reembedParams.get("nBins",20)
    doBatch = reembedParams.get("doBatch",False)
    batchMethod = reembedParams.get("batchMethod","Scanorama")
    batchKey = reembedParams.get("batchKey","")
    scanoramaKnn = reembedParams.get("scanoramaKnn",20)
    scanoramaSigma = reembedParams.get("scanoramaSigma",15)
    scanoramaAlpha = reembedParams.get("scanoramaAlpha",0.1)
    scanoramaBatchSize = reembedParams.get("scanoramaBatchSize",5000)
    bbknnNeighborsWithinBatch = reembedParams.get("bbknnNeighborsWithinBatch",3)
    numPCs = reembedParams.get("numPCs",150)
    pcaSolver = reembedParams.get("pcaSolver","randomized")
    neighborsKnn = reembedParams.get("neighborsKnn",20)
    neighborsMethod = reembedParams.get("neighborsMethod","umap")
    distanceMetric = reembedParams.get("distanceMetric","cosine")
    nnaSAM = reembedParams.get("nnaSAM",50)
    weightModeSAM = reembedParams.get("weightModeSAM","dispersion")
    umapMinDist = reembedParams.get("umapMinDist",0.1)
    logTransform = reembedParams.get("logTransform",False)
    scaleData = reembedParams.get("scaleData",False)
    dataLayer = reembedParams.get("dataLayer","X")
    sumNormalizeCells = reembedParams.get("sumNormalizeCells",False)

    
    if dataLayer == ".raw" and adata.raw is not None:
        adata_raw = AnnData(X=adata.raw.X)
        adata_raw.var_names = adata.raw.var_names
        adata_raw.obs_names = adata.obs_names
        adata_raw.obs = adata.obs
        for key in adata.var.keys():
            adata_raw.var[key] = adata.var[key]
    elif dataLayer == ".raw":
        adata_raw = adata
    elif dataLayer == "X":
        adata_raw = adata
        if dataLayer == "X" and "X" not in adata_raw.layers.keys():
            adata_raw.layers["X"] = adata_raw.X    
        adata_raw.X = adata_raw.layers[dataLayer]        
    else:
        adata_raw = AnnData(X=adata.layers[dataLayer])
        adata_raw.var_names = adata.raw.var_names
        adata_raw.obs_names = adata.obs_names
        adata_raw.obs = adata.obs
        for key in adata.var.keys():
            adata_raw.var[key] = adata.var[key]
                
    if doPreprocess:
        # sc.pp.filter_cells(adata_raw,min_counts=minCountsCF,min_genes=minGenesCF)
        sc.pp.filter_genes(adata_raw, min_counts=minCountsGF)
        sc.pp.filter_genes(adata_raw, min_cells=minCellsGF/100*adata_raw.shape[0])
        sc.pp.filter_genes(adata_raw, max_cells=maxCellsGF/100*adata_raw.shape[0])
        
        if not doSAM:
            batchKey = None if (batchKey == "" or not doBatch) else batchKey
            sc.pp.highly_variable_genes(adata_raw,batch_key=batchKey,flavor='seurat_v3',n_top_genes=nTopGenesHVG, n_bins=nBinsHVG)

    if doSAM:
        SAM = get_samalg_module()
        sam=SAM(counts = adata_raw, inplace=True)
        sn = "cell_median" if sumNormalizeCells else None
        norm = "log" if logTransform else None
        sam.preprocess_data(sum_norm=sn,thresh_low=0,thresh_high=1.0, norm=norm, min_expression=0,filter_genes=False)
        # do yo SAM
    else:
        # do yo scanpy
        if sumNormalizeCells:
            sc.pp.normalize_total(adata_raw)
        if logTransform:
            sc.pp.log1p(adata_raw)


    
    if not doSAM:
        X = adata_raw.X
        if scaleData:
            print('Scaling')
            sc.pp.scale(adata_raw,max_value=10)
        sc.pp.pca(adata_raw,n_comps=min(adata.n_vars - 1, numPCs), svd_solver=pcaSolver)
        adata_raw.X = X
    else:
        X = sam.adata.X
        preprocessing = "StandardScaler" if scaleData else "Normalizer"
        sam.run(projection=None,weight_mode=weightModeSAM,preprocessing=preprocessing,distance=distanceMetric,num_norm_avg=nnaSAM)
        sam.adata.X = X        
        adata_raw=sam.adata

    if doBatch:
        sce = get_scanpy_external_module()
        if doSAM:
            adata_batch = sam.adata
        else:
            adata_batch = adata_raw
        
        if batchMethod == "Harmony":
            sce.pp.harmony_integrate(adata_batch,batchKey,adjusted_basis="X_pca")
        elif batchMethod == "BBKNN":
            metric = "angular" if distanceMetric in ["correlation","cosine"] else "euclidean"
            sce.pp.bbknn(adata_batch, batch_key=batchKey, metric=metric, n_pcs=numPCs, neighbors_within_batch=bbknnNeighborsWithinBatch)
        elif batchMethod == "Scanorama":
            sce.pp.scanorama_integrate(adata_batch, batchKey, basis='X_pca', adjusted_basis='X_pca',
                                    knn=scanoramaKnn, sigma=scanoramaSigma, alpha=scanoramaAlpha,
                                    batch_size=scanoramaBatchSize)
        if doSAM:
            sam.adata = adata_batch
        else:
            adata_raw = adata_batch
    
    if not doSAM or doSAM and batchMethod == "BBKNN":
        if not doBatch or doBatch and batchMethod != "BBKNN":
            sc.pp.neighbors(adata_raw, n_neighbors=neighborsKnn, use_rep="X_pca",method=neighborsMethod, metric=distanceMetric)    
        sc.tl.umap(adata_raw, min_dist=umapMinDist)
    else:
        sam.run_umap(metric=distanceMetric,min_dist=umapMinDist)
        adata_raw.obsm['X_umap'] = sam.adata.obsm['X_umap']
        adata_raw.obsp['connectivities'] = sam.adata.obsp['connectivities']
        
    umap = adata_raw.obsm["X_umap"]
    result = np.full((obs_mask.shape[0], umap.shape[1]), np.NaN)
    result[obs_mask] = umap
    return result, adata_raw.obsp['connectivities']



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: NetEncoding

import flatbuffers

class Matrix(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMatrix(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Matrix()
        x.Init(buf, n + offset)
        return x

    # Matrix
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Matrix
    def NRows(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Matrix
    def NCols(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Matrix
    def Columns(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Column import Column
            obj = Column()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Matrix
    def ColumnsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Matrix
    def ColIndexType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Matrix
    def ColIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Matrix
    def RowIndexType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Matrix
    def RowIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

def MatrixStart(builder): builder.StartObject(7)
def MatrixAddNRows(builder, nRows): builder.PrependUint32Slot(0, nRows, 0)
def MatrixAddNCols(builder, nCols): builder.PrependUint32Slot(1, nCols, 0)
def MatrixAddColumns(builder, columns): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(columns), 0)
def MatrixStartColumnsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def MatrixAddColIndexType(builder, colIndexType): builder.PrependUint8Slot(3, colIndexType, 0)
def MatrixAddColIndex(builder, colIndex): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(colIndex), 0)
def MatrixAddRowIndexType(builder, rowIndexType): builder.PrependUint8Slot(5, rowIndexType, 0)
def MatrixAddRowIndex(builder, rowIndex): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(rowIndex), 0)
def MatrixEnd(builder): return builder.EndObject()



# Work around bug https://github.com/pallets/werkzeug/issues/461
if __package__ is None:
    import sys
    from pathlib import Path

    PKG_PATH = Path(__file__).parent
    sys.path.insert(0, str(PKG_PATH.parent))
    import backend.server # noqa F401
    __package__ = PKG_PATH.name

# Main thing
from .cli.cli import cli  # noqa F402

cli()



import json

import numpy as np
import pandas as pd
from flatbuffers import Builder
from scipy import sparse

import backend.common.fbs.NetEncoding.Column as Column
import backend.common.fbs.NetEncoding.Float32Array as Float32Array
import backend.common.fbs.NetEncoding.Float64Array as Float64Array
import backend.common.fbs.NetEncoding.Int32Array as Int32Array
import backend.common.fbs.NetEncoding.JSONEncodedArray as JSONEncodedArray
import backend.common.fbs.NetEncoding.Matrix as Matrix
import backend.common.fbs.NetEncoding.TypedArray as TypedArray
import backend.common.fbs.NetEncoding.Uint32Array as Uint32Array

# Serialization helper
def serialize_column(builder, typed_arr):
    """ Serialize NetEncoding.Column """

    (u_type, u_value) = typed_arr
    Column.ColumnStart(builder)
    Column.ColumnAddUType(builder, u_type)
    Column.ColumnAddU(builder, u_value)
    return Column.ColumnEnd(builder)


# Serialization helper
def serialize_matrix(builder, n_rows, n_cols, columns, col_idx):
    """ Serialize NetEncoding.Matrix """

    Matrix.MatrixStart(builder)
    Matrix.MatrixAddNRows(builder, n_rows)
    Matrix.MatrixAddNCols(builder, n_cols)
    Matrix.MatrixAddColumns(builder, columns)
    if col_idx is not None:
        (u_type, u_val) = col_idx
        Matrix.MatrixAddColIndexType(builder, u_type)
        Matrix.MatrixAddColIndex(builder, u_val)
    return Matrix.MatrixEnd(builder)


# Serialization helper
def serialize_typed_array(builder, source_array, encoding_info):
    """
    Serialize any of the various typed arrays, eg, Float32Array. Specific means of serialization and type conversion
    are provided by type_info.
    """

    arr = source_array
    (array_type, as_type) = encoding_info(source_array)

    if isinstance(arr, pd.Index):
        arr = arr.to_series()

    # convert to a simple ndarray
    if as_type == "json":
        as_json = arr.to_json(orient="records")
        arr = np.array(bytearray(as_json, "utf-8"))
    else:
        if sparse.issparse(arr):
            arr = arr.toarray()
        elif isinstance(arr, pd.Series):
            arr = arr.to_numpy()
        if arr.dtype != as_type:
            arr = arr.astype(as_type)

    # serialize the ndarray into a vector
    if arr.ndim == 2:
        if arr.shape[0] == 1:
            arr = arr[0]
        elif arr.shape[1] == 1:
            arr = arr.T[0]
    if isinstance(arr,np.matrix):
        arr = arr.A.flatten()
    
    vec = builder.CreateNumpyVector(arr)

    # serialize the typed array table
    builder.StartObject(1)
    builder.PrependUOffsetTRelativeSlot(0, vec, 0)
    array_value = builder.EndObject()
    return (array_type, array_value)


def column_encoding(arr):
    column_encoding_type_map = {
        # array protocol string:  ( array_type, as_type )
        np.dtype(np.float64).str: (TypedArray.TypedArray.Float64Array, np.float64),
        np.dtype(np.float32).str: (TypedArray.TypedArray.Float32Array, np.float32),
        np.dtype(np.float16).str: (TypedArray.TypedArray.Float32Array, np.float32),
        np.dtype(np.int8).str: (TypedArray.TypedArray.Int32Array, np.int32),
        np.dtype(np.int16).str: (TypedArray.TypedArray.Int32Array, np.int32),
        np.dtype(np.int32).str: (TypedArray.TypedArray.Int32Array, np.int32),
        np.dtype(np.int64).str: (TypedArray.TypedArray.Int32Array, np.int32),
        np.dtype(np.uint8).str: (TypedArray.TypedArray.Uint32Array, np.uint32),
        np.dtype(np.uint16).str: (TypedArray.TypedArray.Uint32Array, np.uint32),
        np.dtype(np.uint32).str: (TypedArray.TypedArray.Uint32Array, np.uint32),
        np.dtype(np.uint64).str: (TypedArray.TypedArray.Uint32Array, np.uint32),
    }
    column_encoding_default = (TypedArray.TypedArray.JSONEncodedArray, "json")

    return column_encoding_type_map.get(arr.dtype.str, column_encoding_default)


def index_encoding(arr):
    index_encoding_type_map = {
        # array protocol string:  ( array_type, as_type )
        np.dtype(np.int32).str: (TypedArray.TypedArray.Int32Array, np.int32),
        np.dtype(np.int64).str: (TypedArray.TypedArray.Int32Array, np.int32),
        np.dtype(np.uint32).str: (TypedArray.TypedArray.Uint32Array, np.uint32),
        np.dtype(np.uint64).str: (TypedArray.TypedArray.Uint32Array, np.uint32),
    }
    index_encoding_default = (TypedArray.TypedArray.JSONEncodedArray, "json")

    return index_encoding_type_map.get(arr.dtype.str, index_encoding_default)


def guess_at_mem_needed(matrix):
    (n_rows, n_cols) = matrix.shape
    if isinstance(matrix, np.ndarray) or sparse.issparse(matrix):
        guess = (n_rows * n_cols * matrix.dtype.itemsize) + 1024
    elif isinstance(matrix, pd.DataFrame):
        # XXX TODO - DataFrame type estimate
        guess = 1
    else:
        guess = 1

    # round up to nearest 1024 bytes
    guess = (guess + 0x400) & (~0x3FF)
    return guess


def encode_matrix_fbs(matrix, row_idx=None, col_idx=None):
    """
    Given a 2D DataFrame, ndarray or sparse equivalent, create and return a Matrix flatbuffer.

    :param matrix: 2D DataFrame, ndarray or sparse equivalent
    :param row_idx: index for row dimension, Index or ndarray
    :param col_idx: index for col dimension, Index or ndarray

    NOTE: row indices are (currently) unsupported and must be None
    """

    if row_idx is not None:
        raise ValueError("row indexing not supported for FBS Matrix")
    if matrix.ndim != 2:
        raise ValueError("FBS Matrix must be 2D")

    (n_rows, n_cols) = matrix.shape

    # estimate size needed, so we don't unnecessarily realloc.
    builder = Builder(guess_at_mem_needed(matrix))

    columns = []
    for cidx in range(n_cols - 1, -1, -1):
        # serialize the typed array
        col = matrix.iloc[:, cidx] if isinstance(matrix, pd.DataFrame) else matrix[:, cidx]
        typed_arr = serialize_typed_array(builder, col, column_encoding)

        # serialize the Column union
        columns.append(serialize_column(builder, typed_arr))

    # Serialize Matrix.columns[]
    Matrix.MatrixStartColumnsVector(builder, n_cols)
    for c in columns:
        builder.PrependUOffsetTRelative(c)
    matrix_column_vec = builder.EndVector(n_cols)

    # serialize the colIndex if provided
    cidx = None
    if col_idx is not None:
        cidx = serialize_typed_array(builder, col_idx, index_encoding)

    # Serialize Matrix
    matrix = serialize_matrix(builder, n_rows, n_cols, matrix_column_vec, cidx)

    builder.Finish(matrix)
    return builder.Output()


def deserialize_typed_array(tarr):
    type_map = {
        TypedArray.TypedArray.NONE: None,
        TypedArray.TypedArray.Uint32Array: Uint32Array.Uint32Array,
        TypedArray.TypedArray.Int32Array: Int32Array.Int32Array,
        TypedArray.TypedArray.Float32Array: Float32Array.Float32Array,
        TypedArray.TypedArray.Float64Array: Float64Array.Float64Array,
        TypedArray.TypedArray.JSONEncodedArray: JSONEncodedArray.JSONEncodedArray,
    }
    (u_type, u) = tarr
    if u_type is TypedArray.TypedArray.NONE:
        return None

    TarType = type_map.get(u_type, None)
    if TarType is None:
        raise TypeError(f"FBS contains unknown data type: {u_type}")

    arr = TarType()
    arr.Init(u.Bytes, u.Pos)
    narr = arr.DataAsNumpy()
    if u_type == TypedArray.TypedArray.JSONEncodedArray:
        narr = json.loads(narr.tostring().decode("utf-8"))
    return narr


def decode_matrix_fbs(fbs):
    """
    Given an FBS-encoded Matrix, return a Pandas DataFrame the contains the data and indices.
    """

    matrix = Matrix.Matrix.GetRootAsMatrix(fbs, 0)
    n_rows = matrix.NRows()
    n_cols = matrix.NCols()
    if n_rows == 0 or n_cols == 0:
        return pd.DataFrame()

    if matrix.RowIndexType() is not TypedArray.TypedArray.NONE:
        raise ValueError("row indexing not supported for FBS Matrix")

    columns_length = matrix.ColumnsLength()

    columns_index = deserialize_typed_array((matrix.ColIndexType(), matrix.ColIndex()))
    if columns_index is None:
        columns_index = range(0, n_cols)

    # sanity checks
    if len(columns_index) != n_cols or columns_length != n_cols:
        raise ValueError("FBS column count does not match number of columns in underlying matrix")

    columns_data = {}
    columns_type = {}
    for col_idx in range(0, columns_length):
        col = matrix.Columns(col_idx)
        tarr = (col.UType(), col.U())
        data = deserialize_typed_array(tarr)
        columns_data[columns_index[col_idx]] = data
        if len(data) != n_rows:
            raise ValueError("FBS column length does not match number of rows")
        if col.UType() is TypedArray.TypedArray.JSONEncodedArray:
            columns_type[columns_index[col_idx]] = "category"

    df = pd.DataFrame.from_dict(data=columns_data).astype(columns_type, copy=False)

    # more sanity checks
    if not df.columns.is_unique or len(df.columns) != n_cols:
        raise KeyError("FBS column indices are not unique")

    return df



import os
from os.path import splitext, isdir

from backend.server.common.annotations.local_file_csv import AnnotationsLocalFile
from backend.server.common.config.base_config import BaseConfig
from backend.common.errors import ConfigurationError, OntologyLoadFailure, AnnotationsError
from backend.server.compute.scanpy import get_scanpy_module
from backend.server.data_common.matrix_loader import MatrixDataLoader


class DatasetConfig(BaseConfig):
    """Manages the config attribute associated with a dataset."""

    def __init__(self, tag, app_config, default_config):
        super().__init__(app_config, default_config)
        self.tag = tag
        try:
            self.app__scripts = default_config["app"]["scripts"]
            self.app__inline_scripts = default_config["app"]["inline_scripts"]
            self.app__authentication_enable = default_config["app"]["authentication_enable"]

            self.presentation__max_categories = default_config["presentation"]["max_categories"]
            self.presentation__custom_colors = default_config["presentation"]["custom_colors"]

            self.user_annotations__enable = default_config["user_annotations"]["enable"]
            self.user_annotations__type = default_config["user_annotations"]["type"]
            self.user_annotations__local_file_csv__directory = default_config["user_annotations"]["local_file_csv"][
                "directory"
            ]
            self.user_annotations__local_file_csv__file = default_config["user_annotations"]["local_file_csv"]["file"]
            self.user_annotations__ontology__enable = default_config["user_annotations"]["ontology"]["enable"]
            self.user_annotations__ontology__obo_location = default_config["user_annotations"]["ontology"][
                "obo_location"
            ]
            self.user_annotations__gene_sets__readonly = default_config["user_annotations"]["gene_sets"]["readonly"]
            self.user_annotations__local_file_csv__gene_sets_file = default_config["user_annotations"][
                "local_file_csv"
            ]["gene_sets_file"]
            self.user_annotations__local_file_csv__reembed_parameters_file = default_config["user_annotations"][
                "local_file_csv"
            ]["reembed_parameters_file"]
            self.embeddings__names = default_config["embeddings"]["names"]

            self.diffexp__enable = default_config["diffexp"]["enable"]
            self.diffexp__lfc_cutoff = default_config["diffexp"]["lfc_cutoff"]
            self.diffexp__top_n = default_config["diffexp"]["top_n"]

        except KeyError as e:
            raise ConfigurationError(f"Unexpected config: {str(e)}")

        # The annotation object is created during complete_config and stored here.
        self.user_annotations = None

    def complete_config(self, context):
        self.handle_app()
        self.handle_presentation()
        self.handle_user_annotations(context)
        self.handle_embeddings()
        self.handle_diffexp(context)

    def get_data_adaptor(self):
        server_config = self.app_config.server_config
        if not server_config.data_adaptor:
            matrix_data_loader = MatrixDataLoader(server_config.single_dataset__datapath, app_config=self.app_config)
            server_config.data_adaptor = matrix_data_loader.open(self.app_config)

        return server_config.data_adaptor

    def handle_app(self):
        self.validate_correct_type_of_configuration_attribute("app__scripts", list)
        self.validate_correct_type_of_configuration_attribute("app__inline_scripts", list)
        self.validate_correct_type_of_configuration_attribute("app__authentication_enable", bool)

        # scripts can be string (filename) or dict (attributes). Convert string to dict.
        scripts = []
        for script in self.app__scripts:
            try:
                if isinstance(script, str):
                    scripts.append({"src": script})
                elif isinstance(script, dict) and isinstance(script["src"], str):
                    scripts.append(script)
                else:
                    raise Exception
            except Exception as e:
                raise ConfigurationError(f"Scripts must be string or a dict containing an src key: {e}")

        self.app__scripts = scripts

    def handle_presentation(self):
        self.validate_correct_type_of_configuration_attribute("presentation__max_categories", int)
        self.validate_correct_type_of_configuration_attribute("presentation__custom_colors", bool)

    def handle_user_annotations(self, context):
        self.validate_correct_type_of_configuration_attribute("user_annotations__enable", bool)
        self.validate_correct_type_of_configuration_attribute("user_annotations__type", str)
        self.validate_correct_type_of_configuration_attribute(
            "user_annotations__local_file_csv__directory", (type(None), str)
        )
        self.validate_correct_type_of_configuration_attribute(
            "user_annotations__local_file_csv__file", (type(None), str)
        )
        self.validate_correct_type_of_configuration_attribute(
            "user_annotations__local_file_csv__gene_sets_file", (type(None), str)
        )
        self.validate_correct_type_of_configuration_attribute(
            "user_annotations__local_file_csv__reembed_parameters_file", (type(None), str)
        )        
        self.validate_correct_type_of_configuration_attribute("user_annotations__ontology__enable", bool)
        self.validate_correct_type_of_configuration_attribute(
            "user_annotations__ontology__obo_location", (type(None), str)
        )
        self.validate_correct_type_of_configuration_attribute("user_annotations__gene_sets__readonly", bool)

        if self.user_annotations__enable or not self.user_annotations__gene_sets__readonly:
            server_config = self.app_config.server_config
            if not self.app__authentication_enable:
                raise ConfigurationError("user annotations requires authentication to be enabled")
            if not server_config.auth.is_valid_authentication_type():
                auth_type = server_config.authentication__type
                raise ConfigurationError(f"authentication method {auth_type} is not compatible with user annotations")

        # Must always have an annotations instance to support genesets. User annotation (cell labels) are optional
        # as are writable gene sets
        if self.user_annotations__type == "local_file_csv":
            self.handle_local_file_csv_annotations(context)
        else:
            raise ConfigurationError('The only annotation type support is "local_file_csv"')

        if self.user_annotations__enable:
            if self.user_annotations__ontology__enable or self.user_annotations__ontology__obo_location:
                try:
                    self.user_annotations.load_ontology(self.user_annotations__ontology__obo_location)
                except OntologyLoadFailure as e:
                    raise ConfigurationError("Unable to load ontology terms\n" + str(e))

        self.check_annotation_config_vars_not_set(context)

    def handle_local_file_csv_annotations(self, context):
        dirname = self.user_annotations__local_file_csv__directory
        filename = self.user_annotations__local_file_csv__file
        genesets_filename = self.user_annotations__local_file_csv__gene_sets_file
        reembed_parameters_filename = self.user_annotations__local_file_csv__reembed_parameters_file

        if dirname is not None and (filename is not None or genesets_filename is not None):
            raise ConfigurationError(
                "'user-generated-data-dir' may not be used with 'annotations-file' or 'gene-sets-file'."
            )

        if filename is not None:
            lf_name, lf_ext = splitext(filename)
            if lf_ext and lf_ext != ".csv":
                raise ConfigurationError(f"annotation file type must be .csv: {filename}")

        if genesets_filename is not None:
            lf_name, lf_ext = splitext(genesets_filename)
            if lf_ext and lf_ext != ".csv":
                raise ConfigurationError(f"genesets file type must be .csv: {genesets_filename}")
        
        if reembed_parameters_filename is not None:
            lf_name, lf_ext = splitext(reembed_parameters_filename)
            if lf_ext and lf_ext != ".json":
                raise ConfigurationError(f"reembed parameters file type must be .json: {reembed_parameters_filename}")
        
        if dirname is not None and not isdir(dirname):
            try:
                os.mkdir(dirname)
            except OSError:
                raise ConfigurationError("Unable to create directory specified by --user-generated-data-dir")

        anno_config = {
            "user-annotations": self.user_annotations__enable,
            "genesets-save": not self.user_annotations__gene_sets__readonly,
        }
        self.user_annotations = AnnotationsLocalFile(anno_config, dirname, filename, genesets_filename, reembed_parameters_filename)

        # if the user has specified a fixed label file, go ahead and validate it
        # so that we can remove errors early in the process.
        server_config = self.app_config.server_config
        if server_config.single_dataset__datapath:
            data_adaptor = self.get_data_adaptor()
            if self.user_annotations__local_file_csv__file:
                self.user_annotations.read_labels(data_adaptor)
            if self.user_annotations__local_file_csv__gene_sets_file:
                try:
                    self.user_annotations.read_gene_sets(data_adaptor, context)
                except (ValueError, AnnotationsError, KeyError) as e:
                    raise ConfigurationError(f"Unable to read genesets CSV file: {str(e)}") from e
            if self.user_annotations__local_file_csv__reembed_parameters_file:
                try:
                    self.user_annotations.read_reembed_parameters(data_adaptor, context)
                except (ValueError, AnnotationsError, KeyError) as e:
                    raise ConfigurationError(f"Unable to read reembed JSON file: {str(e)}") from e                    

    def check_annotation_config_vars_not_set(self, context):
        if self.user_annotations__type is not None:
            dirname = self.user_annotations__local_file_csv__directory
            filename = self.user_annotations__local_file_csv__file
            if not self.user_annotations__enable:
                if filename is not None:
                    context["messagefn"]("Warning: --annotations-file ignored as annotations are disabled.")
                if self.user_annotations__ontology__enable:
                    context["messagefn"](
                        "Warning: --experimental-annotations-ontology ignored as annotations are disabled."
                    )
                if self.user_annotations__ontology__obo_location is not None:
                    context["messagefn"](
                        "Warning: --experimental-annotations-ontology-obo ignored as annotations are disabled."
                    )
                if dirname is not None:
                    context["messagefn"]("Warning: --user-generated-data-dir ignored as annotations are disabled.")

    def handle_embeddings(self):
        self.validate_correct_type_of_configuration_attribute("embeddings__names", list)

    def handle_diffexp(self, context):
        self.validate_correct_type_of_configuration_attribute("diffexp__enable", bool)
        self.validate_correct_type_of_configuration_attribute("diffexp__lfc_cutoff", float)
        self.validate_correct_type_of_configuration_attribute("diffexp__top_n", int)

        data_adaptor = self.get_data_adaptor()
        if self.diffexp__enable and data_adaptor.parameters.get("diffexp_may_be_slow", False):
            context["messagefn"](
                "CAUTION: due to the size of your dataset, " "running differential expression may take longer or fail."
            )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: NetEncoding

import flatbuffers

class Float64Array(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsFloat64Array(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Float64Array()
        x.Init(buf, n + offset)
        return x

    # Float64Array
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Float64Array
    def Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Float64Array
    def DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Float64Array
    def DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def Float64ArrayStart(builder): builder.StartObject(1)
def Float64ArrayAddData(builder, data): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0)
def Float64ArrayStartDataVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def Float64ArrayEnd(builder): return builder.EndObject()



from abc import ABCMeta, abstractmethod

import fastobo
import fsspec

from backend.common.errors import OntologyLoadFailure, DisabledFeatureError
from backend.common.utils.type_conversion_utils import get_schema_type_hint_of_array
from backend.common.genesets import write_gene_sets_tidycsv


class Annotations(metaclass=ABCMeta):
    """ baseclass for annotations, including ontologies, gene sets, and re-embedding parameters """

    """ our default ontology is the PURL for the Cell Ontology.
    See http://www.obofoundry.org/ontology/cl.html """
    DefaultOnotology = "http://purl.obolibrary.org/obo/cl.obo"

    def __init__(self, config={}):
        self.ontology_data = None
        self.config = config

    def user_annotations_enabled(self):
        return self.config.get("user-annotations", False)

    def gene_sets_save_enabled(self):
        return self.config.get("genesets-save", False)

    def check_user_annotations_enabled(self):
        if not self.user_annotations_enabled():
            raise DisabledFeatureError("User annotations are disabled.")

    def check_gene_sets_save_enabled(self):
        if not self.gene_sets_save_enabled():
            raise DisabledFeatureError("User gene sets save is disabled.")

    def load_ontology(self, path):
        """Load and parse ontologies - currently support OBO files only."""
        if path is None:
            path = self.DefaultOnotology

        try:
            with fsspec.open(path) as f:
                obo = fastobo.iter(f)
                terms = filter(lambda stanza: type(stanza) is fastobo.term.TermFrame, obo)
                names = [tag.name for term in terms for tag in term if type(tag) is fastobo.term.NameClause]
                self.ontology_data = names

        except FileNotFoundError as e:
            raise OntologyLoadFailure("Unable to find OBO ontology path") from e

        except SyntaxError as e:
            raise OntologyLoadFailure("Syntax error loading OBO ontology") from e

        except Exception as e:
            raise OntologyLoadFailure("Error loading OBO file") from e

    def get_schema(self, data_adaptor):
        schema = []
        labels = self.read_labels(data_adaptor)
        if labels is not None and not labels.empty:
            for col in labels.columns:
                col_schema = dict(name=col, writable=True)
                col_schema.update(get_schema_type_hint_of_array(labels[col]))
                schema.append(col_schema)

        return schema

    @abstractmethod
    def set_collection(self, name):
        """set or create a new annotation collection"""
        pass

    @abstractmethod
    def read_labels(self, data_adaptor):
        """Return the labels as a pandas.DataFrame"""
        pass

    @abstractmethod
    def write_labels(self, df, data_adaptor):
        """Write the labels (df) to a persistent storage such that it can later be read"""
        pass

    @abstractmethod
    def read_gene_sets(self, data_adaptor):
        """Return the gene sets from persistent storage """
        pass

    @abstractmethod
    def read_reembed_parameters(self, data_adaptor):
        """Return the reembedding parameters from persistent storage """
        pass        

    @abstractmethod
    def write_gene_sets(self, gs, data_adaptor):
        """Write the gene sets (gs) to a persistent storage such that it can later be read"""
        pass

    @abstractmethod
    def write_reembed_parameters(self, parameters, data_adaptor):
        """Write the reembedding parameters (parameters) to a persistent storage such that it can later be read"""
        pass

    @abstractmethod
    def update_parameters(self, parameters, data_adaptor):
        """Update configuration parameters that describe information about the annotations feature"""
        pass

    @staticmethod
    def gene_sets_to_csv(genesets):
        """
        Convert the internal gene sets format (returned by read_gene_set) into
        the simple Tidy CSV.
        """
        from io import StringIO

        if isinstance(genesets, dict):
            genesets = genesets.values()

        with StringIO() as sio:
            write_gene_sets_tidycsv(sio, genesets)
            return sio.getvalue()

    @staticmethod
    def gene_sets_to_response(genesets):
        """
        Convert the internal gene sets format (returned by read_gene_set) into
        the dict expected by the JSON REST API
        """
        return list(genesets.values())



from backend.server import display_version as cellxgene_display_version


def get_client_config(app_config, data_adaptor):
    """
    Return the configuration as required by the /config REST route
    """

    server_config = app_config.server_config
    dataset_config = data_adaptor.dataset_config
    annotation = dataset_config.user_annotations
    auth = server_config.auth

    # FIXME The current set of config is not consistently presented:
    # we have camalCase, hyphen-text, and underscore_text

    # make sure the configuration has been checked.
    app_config.check_config()

    # display_names
    title = app_config.get_title(data_adaptor)
    about = app_config.get_about(data_adaptor)

    display_names = dict(engine=data_adaptor.get_name(), dataset=title)

    # library_versions
    library_versions = {}
    library_versions.update(data_adaptor.get_library_versions())
    library_versions["cellxgene"] = cellxgene_display_version

    # links
    links = {"about-dataset": about}

    # parameters
    parameters = {
        "layout": dataset_config.embeddings__names,
        "max-category-items": dataset_config.presentation__max_categories,
        "obs_names": server_config.single_dataset__obs_names,
        "var_names": server_config.single_dataset__var_names,
        "diffexp_lfc_cutoff": dataset_config.diffexp__lfc_cutoff,
        "backed": server_config.adaptor__anndata_adaptor__backed,
        "disable-diffexp": not dataset_config.diffexp__enable,
        "annotations": False,
        "annotations_file": None,
        "annotations_dir": None,
        "annotations_genesets": True,  # feature flag
        "annotations_genesets_readonly": dataset_config.user_annotations__gene_sets__readonly,
        "annotations_genesets_summary_methods": ["mean"],
        "annotations_cell_ontology_enabled": False,
        "annotations_cell_ontology_obopath": None,
        "annotations_cell_ontology_terms": None,
        "custom_colors": dataset_config.presentation__custom_colors,
        "diffexp-may-be-slow": False,
    }

    # corpora dataset_props
    # TODO/Note: putting info from the dataset into the /config is not ideal.
    # However, it is definitely not part of /schema, and we do not have a top-level
    # route for data properties.  Consider creating one at some point.
    corpora_props = data_adaptor.get_corpora_props()
    if corpora_props and "default_embedding" in corpora_props:
        default_embedding = corpora_props["default_embedding"]
        if isinstance(default_embedding, str) and default_embedding.startswith("X_"):
            default_embedding = default_embedding[2:]  # drop X_ prefix
        if default_embedding in data_adaptor.get_embedding_names():
            parameters["default_embedding"] = default_embedding

    data_adaptor.update_parameters(parameters)
    if annotation:
        annotation.update_parameters(parameters, data_adaptor)

    # gather it all together
    client_config = {}
    config = client_config["config"] = {}
    config["displayNames"] = display_names
    config["library_versions"] = library_versions
    config["links"] = links
    config["parameters"] = parameters
    config["corpora_props"] = corpora_props
    config["limits"] = {
        "column_request_max": server_config.limits__column_request_max,
        "diffexp_cellcount_max": server_config.limits__diffexp_cellcount_max,
    }

    if dataset_config.app__authentication_enable and auth.is_valid_authentication_type():
        config["authentication"] = {
            "requires_client_login": auth.requires_client_login(),
        }
        if auth.requires_client_login():
            config["authentication"].update(
                {
                    # Todo why are these stored on the data_adaptor?
                    "login": auth.get_login_url(data_adaptor),
                    "logout": auth.get_logout_url(data_adaptor),
                }
            )

    return client_config


def get_client_userinfo(app_config, data_adaptor):
    """
    Return the userinfo as required by the /userinfo REST route
    """

    server_config = app_config.server_config
    dataset_config = data_adaptor.dataset_config
    auth = server_config.auth

    # make sure the configuration has been checked.
    app_config.check_config()

    if dataset_config.app__authentication_enable and auth.is_valid_authentication_type():
        userinfo = {}
        userinfo["userinfo"] = {
            "is_authenticated": auth.is_user_authenticated(),
            "username": auth.get_user_name(),
            "user_id": auth.get_user_id(),
            "email": auth.get_user_email(),
            "picture": auth.get_user_picture(),
        }
        return userinfo



import re

from backend.common.errors import ColorFormatException

HEX_COLOR_FORMAT = re.compile("^#[a-fA-F0-9]{6,6}$")

# https://www.w3.org/TR/css-color-4/#named-colors
CSS4_NAMED_COLORS = dict(
    aliceblue="#f0f8ff",
    antiquewhite="#faebd7",
    aqua="#00ffff",
    aquamarine="#7fffd4",
    azure="#f0ffff",
    beige="#f5f5dc",
    bisque="#ffe4c4",
    black="#000000",
    blanchedalmond="#ffebcd",
    blue="#0000ff",
    blueviolet="#8a2be2",
    brown="#a52a2a",
    burlywood="#deb887",
    cadetblue="#5f9ea0",
    chartreuse="#7fff00",
    chocolate="#d2691e",
    coral="#ff7f50",
    cornflowerblue="#6495ed",
    cornsilk="#fff8dc",
    crimson="#dc143c",
    cyan="#00ffff",
    darkblue="#00008b",
    darkcyan="#008b8b",
    darkgoldenrod="#b8860b",
    darkgray="#a9a9a9",
    darkgreen="#006400",
    darkgrey="#a9a9a9",
    darkkhaki="#bdb76b",
    darkmagenta="#8b008b",
    darkolivegreen="#556b2f",
    darkorange="#ff8c00",
    darkorchid="#9932cc",
    darkred="#8b0000",
    darksalmon="#e9967a",
    darkseagreen="#8fbc8f",
    darkslateblue="#483d8b",
    darkslategray="#2f4f4f",
    darkslategrey="#2f4f4f",
    darkturquoise="#00ced1",
    darkviolet="#9400d3",
    deeppink="#ff1493",
    deepskyblue="#00bfff",
    dimgray="#696969",
    dimgrey="#696969",
    dodgerblue="#1e90ff",
    firebrick="#b22222",
    floralwhite="#fffaf0",
    forestgreen="#228b22",
    fuchsia="#ff00ff",
    gainsboro="#dcdcdc",
    ghostwhite="#f8f8ff",
    gold="#ffd700",
    goldenrod="#daa520",
    gray="#808080",
    green="#008000",
    greenyellow="#adff2f",
    grey="#808080",
    honeydew="#f0fff0",
    hotpink="#ff69b4",
    indianred="#cd5c5c",
    indigo="#4b0082",
    ivory="#fffff0",
    khaki="#f0e68c",
    lavender="#e6e6fa",
    lavenderblush="#fff0f5",
    lawngreen="#7cfc00",
    lemonchiffon="#fffacd",
    lightblue="#add8e6",
    lightcoral="#f08080",
    lightcyan="#e0ffff",
    lightgoldenrodyellow="#fafad2",
    lightgray="#d3d3d3",
    lightgreen="#90ee90",
    lightgrey="#d3d3d3",
    lightpink="#ffb6c1",
    lightsalmon="#ffa07a",
    lightseagreen="#20b2aa",
    lightskyblue="#87cefa",
    lightslategray="#778899",
    lightslategrey="#778899",
    lightsteelblue="#b0c4de",
    lightyellow="#ffffe0",
    lime="#00ff00",
    limegreen="#32cd32",
    linen="#faf0e6",
    magenta="#ff00ff",
    maroon="#800000",
    mediumaquamarine="#66cdaa",
    mediumblue="#0000cd",
    mediumorchid="#ba55d3",
    mediumpurple="#9370db",
    mediumseagreen="#3cb371",
    mediumslateblue="#7b68ee",
    mediumspringgreen="#00fa9a",
    mediumturquoise="#48d1cc",
    mediumvioletred="#c71585",
    midnightblue="#191970",
    mintcream="#f5fffa",
    mistyrose="#ffe4e1",
    moccasin="#ffe4b5",
    navajowhite="#ffdead",
    navy="#000080",
    oldlace="#fdf5e6",
    olive="#808000",
    olivedrab="#6b8e23",
    orange="#ffa500",
    orangered="#ff4500",
    orchid="#da70d6",
    palegoldenrod="#eee8aa",
    palegreen="#98fb98",
    paleturquoise="#afeeee",
    palevioletred="#db7093",
    papayawhip="#ffefd5",
    peachpuff="#ffdab9",
    peru="#cd853f",
    pink="#ffc0cb",
    plum="#dda0dd",
    powderblue="#b0e0e6",
    purple="#800080",
    rebeccapurple="#663399",
    red="#ff0000",
    rosybrown="#bc8f8f",
    royalblue="#4169e1",
    saddlebrown="#8b4513",
    salmon="#fa8072",
    sandybrown="#f4a460",
    seagreen="#2e8b57",
    seashell="#fff5ee",
    sienna="#a0522d",
    silver="#c0c0c0",
    skyblue="#87ceeb",
    slateblue="#6a5acd",
    slategray="#708090",
    slategrey="#708090",
    snow="#fffafa",
    springgreen="#00ff7f",
    steelblue="#4682b4",
    tan="#d2b48c",
    teal="#008080",
    thistle="#d8bfd8",
    tomato="#ff6347",
    turquoise="#40e0d0",
    violet="#ee82ee",
    wheat="#f5deb3",
    white="#ffffff",
    whitesmoke="#f5f5f5",
    yellow="#ffff00",
    yellowgreen="#9acd32",
)


def convert_color_to_hex_format(unknown):
    """
    Try to convert color info to a hex triplet string https://en.wikipedia.org/wiki/Web_colors#Hex_triplet.

    The function accepts for the following formats:
    - A CSS4 color name, as supported by matplotlib https://matplotlib.org/3.1.0/gallery/color/named_colors.html
    - RGB tuple/list with values ranging from 0.0 to 1.0, as in [0.5, 0.75, 1.0]
    - RFB tuple/list with values ranging from 0 to 255, as in [128, 192, 255]
    - Hex triplet string, as in "#08c0ff"

    :param unknown: color info of unknown format
    :return: a hex triplet representing that color
    """
    try:
        if type(unknown) in (list, tuple) and len(unknown) == 3:
            if all(0.0 <= ele <= 1.0 for ele in unknown):
                tup = tuple(int(ele * 255) for ele in unknown)
            elif all(0 <= ele <= 255 and isinstance(ele, int) for ele in unknown):
                tup = tuple(unknown)
            else:
                raise ColorFormatException("Unknown color iterable format!")
            return "#%02x%02x%02x" % tup
        elif isinstance(unknown, str) and unknown.lower() in CSS4_NAMED_COLORS:
            return CSS4_NAMED_COLORS[unknown.lower()]
        elif isinstance(unknown, str) and HEX_COLOR_FORMAT.match(unknown):
            return unknown.lower()
        else:
            raise ColorFormatException("Unknown color format type!")
    except Exception as e:
        raise ColorFormatException(e)


def convert_anndata_category_colors_to_cxg_category_colors(data):
    """
    Convert color information from anndata files to the excellxgene color data format as described below:
        {
            "<category_name>": {
                "<label_name>": "<color_hex_code>",
                ...
            },
            ...
        }

    For more on the cxg color data structure, see https://github.com/chanzuckerberg/cellxgene/issues/1307.

    For more on the anndata color data structure, see
    https://github.com/chanzuckerberg/cellxgene/issues/1152#issuecomment-587276178.

    Handling of malformed data:
    - For any color info in a adata.uns[f"{category}_colors"] color array that convert_color_to_hex_format cannot
      convert to a hex triplet string, a ColorFormatException is raised
    - No category_name key group is returned for adata.uns[f"{category}_colors"] keys for which there is no
      adata.obs[f"{category}"] key

    :param data: the anndata file
    :return: cellxgene color data structure as described above
    """
    cxg_colors = dict()
    color_key_suffix = "_colors"
    for uns_key in data.uns.keys():
        # find uns array that describes colors for a category
        if not uns_key.endswith(color_key_suffix):
            continue

        # check to see if we actually have observations for that category
        category_name = uns_key[: -len(color_key_suffix)]
        if category_name not in data.obs.keys():
            continue

        # create the cellxgene color entry for this category
        cxg_colors[category_name] = dict(
            zip(data.obs[category_name].cat.categories, [convert_color_to_hex_format(c) for c in data.uns[uns_key]])
        )
    return cxg_colors



from abc import ABC, abstractmethod


class AuthTypeBase(ABC):
    """Base type for all authentication types."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def is_valid_authentication_type(self):
        """Return True if the auth type is valid, e.g. it can return userinfo and username.
          (AuthTypeNone is the only one type that returns False)"""
        pass

    def requires_client_login(self):
        """Return True if the user needs to login from the client (e.g. Login button is shown)"""
        return False

    @abstractmethod
    def complete_setup(self, app):
        """complete any setup that may be needed by this auth type.  The Flask app is passed in.
        This is the last auth function called before the server starts to run."""
        pass

    @abstractmethod
    def is_user_authenticated(self):
        """Return True if the user is authenticated"""
        pass

    @abstractmethod
    def get_user_id(self):
        """Return the id for this user (string)"""
        pass

    @abstractmethod
    def get_user_name(self):
        """Return the name of the user (string)"""
        pass

    @abstractmethod
    def get_user_email(self):
        """Return the name of the user (string)"""
        pass

    def get_user_picture(self):
        """Return the location to the user's picture"""
        return None


class AuthTypeClientBase(AuthTypeBase):
    """Base type for all authentication types that require the client to login"""

    def __init__(self):
        super().__init__()

    def requires_client_login(self):
        return True

    @abstractmethod
    def add_url_rules(self, selfapp):
        """Add url rules to the app (like /login, /logout, etc)"""
        pass

    @abstractmethod
    def get_login_url(self, data_adaptor):
        """Return the url for the login route"""
        pass

    @abstractmethod
    def get_logout_url(self, data_adaptor):
        """Return the url for the logout route"""
        pass


class AuthTypeFactory:
    """Factory class to create an authentication type"""

    auth_types = {}

    @staticmethod
    def register(name, auth_type):
        assert issubclass(auth_type, AuthTypeBase)
        AuthTypeFactory.auth_types[name] = auth_type

    @staticmethod
    def create(name, app_config):
        auth_type = AuthTypeFactory.auth_types.get(name)
        if auth_type is None:
            return None
        return auth_type(app_config)



import copy

from flatten_dict import flatten
from backend.common.errors import ConfigurationError


class BaseConfig(object):
    """
    This class handles the mechanics of updating and checking attributes.
    Derived classes are expected to store the actual attributes
    Currently DatasetConfig and ServerConfig both inherit from BaseConfig.
    """

    def __init__(self, app_config, default_config):
        # reference back to the app_config
        self.app_config = app_config
        # the complete set of attributes and their default values (unflattened)
        self.default_config = default_config
        # used to make sure every attribute value is checked
        self.attr_checked = {key_name: False for key_name in self.create_mapping(default_config).keys()}

    def create_mapping(self, config):
        """
        Create a dictionary where the keys are the name of attributes (using double underscore convention)
        For example: authentication__type

        The values are a tuple,
        - the first item of the tuple is a tuple of path elements (location in config 'tree')
        - the second item is the value of the config parameter

        For example: (('authentication', 'type'), 'session'))
        """
        config_copy = copy.deepcopy(config)
        mapping = {}

        flat_config = flatten(config_copy)
        for key, value in flat_config.items():
            # name of the attribute
            attr = "__".join(key)
            mapping[attr] = (key, value)

        return mapping

    def validate_correct_type_of_configuration_attribute(self, attrname, vtype):
        val = getattr(self, attrname)
        if type(vtype) in (list, tuple):
            if type(val) not in vtype:
                tnames = ",".join([x.__name__ for x in vtype])
                raise ConfigurationError(
                    f"Invalid type for attribute: {attrname}, expected types ({tnames}), got {type(val).__name__}"
                )
        else:
            if type(val) != vtype:
                raise ConfigurationError(
                    f"Invalid type for attribute: {attrname}, "
                    f"expected type {vtype.__name__}, got {type(val).__name__}"
                )

        self.attr_checked[attrname] = True

    def check_config(self):
        mapping = self.create_mapping(self.default_config)
        for key in mapping.keys():
            if not self.attr_checked[key]:
                raise ConfigurationError(f"The attr '{key}' has not been checked")

    def update(self, **kw):
        """Update the attributes defined in kw with their new values."""
        for key, value in kw.items():
            if not hasattr(self, key):
                raise ConfigurationError(f"unknown config parameter {key}.")
            try:
                if type(value) == tuple:
                    # convert tuple values to list values
                    value = list(value)
                setattr(self, key, value)
            except KeyError:
                raise ConfigurationError(f"Unable to set config parameter {key}.")

            self.attr_checked[key] = False

    def update_from_config(self, config, prefix):
        mapping = self.create_mapping(config)
        for attr, (key, value) in mapping.items():
            if not hasattr(self, attr):
                raise ConfigurationError(f"Unknown key from config file: {prefix}__{attr}")
            setattr(self, attr, value)

            self.attr_checked[attr] = False

    def changes_from_default(self):
        """Return all the attribute that are different from the default"""
        mapping = self.create_mapping(self.default_config)
        diff = []
        for attrname, (key, defval) in mapping.items():
            curval = getattr(self, attrname)
            if curval != defval:
                diff.append((attrname, curval, defval))
        return diff



import click

from .launch import launch
from .prepare import prepare
from .upgrade import log_upgrade_check
from .schema import schema_cli
from .. import __version__


@click.group(
    name="excellxgene",
    subcommand_metavar="COMMAND <args>",
    options_metavar="<options>",
    context_settings=dict(max_content_width=85, help_option_names=["-h", "--help"]),
)
@click.help_option("--help", "-h", help="Show this message and exit.")
@click.version_option(
    version=__version__,
    prog_name="excellxgene",
    message="[%(prog)s] Version %(version)s",
    help="Show the software version and exit.",
)
@click.option(
    "--upgrade-check/--no-upgrade-check", default=True, show_default=True, help="Check for release upgrades on start.",
)
def cli(upgrade_check):
    if upgrade_check:
        log_upgrade_check()


cli.add_command(launch)
cli.add_command(prepare)
cli.add_command(schema_cli)



import copy
import logging
import sys
from http import HTTPStatus
import zlib
import json
import pandas as pd
import numpy as np
from glob import glob
from flask import make_response, jsonify, current_app, abort, send_file, after_this_request, session
from werkzeug.urls import url_unquote
from anndata import AnnData
import pickle
from backend.common.utils.type_conversion_utils import get_schema_type_hint_of_array
from backend.server.common.config.client_config import get_client_config, get_client_userinfo
from backend.common.constants import Axis, DiffExpMode, JSON_NaN_to_num_warning_msg
from backend.common.errors import (
    FilterError,
    JSONEncodingValueError,
    PrepareError,
    DisabledFeatureError,
    ExceedsLimitError,
    DatasetAccessError,
    ColorFormatException,
    AnnotationsError,
    ObsoleteRequest,
    UnsupportedSummaryMethod,
)
from backend.common.genesets import summarizeQueryHash
from backend.common.fbs.matrix import decode_matrix_fbs
import os 
import pathlib

def abort_and_log(code, logmsg, loglevel=logging.DEBUG, include_exc_info=False):
    """
    Log the message, then abort with HTTP code. If include_exc_info is true,
    also include current exception via sys.exc_info().
    """
    if include_exc_info:
        exc_info = sys.exc_info()
    else:
        exc_info = False
    current_app.logger.log(loglevel, logmsg, exc_info=exc_info)
    # Do NOT send log message to HTTP response.
    return abort(code)


def _query_parameter_to_filter(args):
    """
    Convert an annotation value filter, if present in the query args,
    into the standard dict filter format used by internal code.

    Query param filters look like:  <axis>:name=value, where value
    may be one of:
        - a range, min,max, where either may be an open range by using an asterisk, eg, 10,*
        - a value
    Eg,
        ...?tissue=lung&obs:tissue=heart&obs:num_reads=1000,*
    """
    filters = {
        "obs": {},
        "var": {},
    }

    # args has already been url-unquoted once.  We assume double escaping
    # on name and value.
    try:
        for key, value in args.items(multi=True):
            axis, name = key.split(":")
            if axis not in ("obs", "var"):
                raise FilterError("unknown filter axis")
            name = url_unquote(name)
            current = filters[axis].setdefault(name, {"name": name})

            val_split = value.split(",")
            if len(val_split) == 1:
                if "min" in current or "max" in current:
                    raise FilterError("do not mix range and value filters")
                value = url_unquote(value)
                values = current.setdefault("values", [])
                values.append(value)

            elif len(val_split) == 2:
                if len(current) > 1:
                    raise FilterError("duplicate range specification")
                min = url_unquote(val_split[0])
                max = url_unquote(val_split[1])
                if min != "*":
                    current["min"] = float(min)
                if max != "*":
                    current["max"] = float(max)
                if len(current) < 2:
                    raise FilterError("must specify at least min or max in range filter")

            else:
                raise FilterError("badly formated filter value")

    except ValueError as e:
        raise FilterError(str(e))

    result = {}
    for axis in ("obs", "var"):
        axis_filter = filters[axis]
        if len(axis_filter) > 0:
            result[axis] = {"annotation_value": [val for val in axis_filter.values()]}

    return result

def schema_get_helper(data_adaptor):
    if data_adaptor.data.raw is not None:
        layers = [".raw"]+list(data_adaptor.data.layers.keys())
    else:
        layers = list(data_adaptor.data.layers.keys())
    
    if "X" not in layers:
        layers = ["X"] + layers
    
    latent_spaces = []
    for k in data_adaptor._obsm_init.keys():
        if data_adaptor._obsm_init[k].shape[1] > 2:
            latent_spaces.append(k)
    
    var_keys = []
    for v in list(data_adaptor.data.var.keys()):
        if ';;tMean' not in v and v != 'name_0':
            var_keys.append(v)

    schema = {
        "dataframe": {"nObs": data_adaptor.cell_count, "nVar": data_adaptor.gene_count, "type": str(data_adaptor.data.X.dtype)},
        "annotations": {
            "obs": {"index": data_adaptor.parameters.get("obs_names"), "columns": []},
            "var": {"index": data_adaptor.parameters.get("var_names"), "columns": []},
        },
        "layout": {"obs": []},
        "layers": layers,
        "var_keys": var_keys,
        "latent_spaces": latent_spaces
    }
    userID = _get_user_id(data_adaptor)   
    
    for ax in Axis:
        if str(ax) == "var":
            curr_axis = getattr(data_adaptor.data, "var")
            for ann in curr_axis:
                ann_schema = {"name": ann, "writable": True}
                ann_schema.update(get_schema_type_hint_of_array(curr_axis[ann]))
                if ann_schema['type']!='categorical':
                    ann_schema['writable']=False
                schema["annotations"]["var"]["columns"].append(ann_schema)
        elif str(ax) == "obs":
            curr_axis = pickle.load(open(f"{userID}/obs.p","rb"))
            for ann in curr_axis:
                ann_schema = {"name": ann, "writable": True}
                ann_schema.update(get_schema_type_hint_of_array(curr_axis[ann]))
                if ann_schema['type']!='categorical':
                    ann_schema['writable']=False
                schema["annotations"][ax]["columns"].append(ann_schema)
            
            ann = "name_0"
            curr_axis = data_adaptor.data.obs
            ann_schema = {"name": ann, "writable": True}
            ann_schema.update(get_schema_type_hint_of_array(curr_axis[ann]))
            if ann_schema['type']!='categorical':
                ann_schema['writable']=False
            schema["annotations"][ax]["columns"].append(ann_schema)

    
    for layout in [x.split(f"{userID}/emb/")[-1][:-2] for x in glob(f"{userID}/emb/*.p")]:
        layout_schema = {"name": layout, "type": "float32", "dims": [f"{layout}_0", f"{layout}_1"]}
        schema["layout"]["obs"].append(layout_schema)
    return schema

def schema_get(data_adaptor):
    schema = schema_get_helper(data_adaptor)
    return make_response(jsonify({"schema": schema}), HTTPStatus.OK)

def gene_info_get(request,data_adaptor):
    gene = request.args.get("gene", None)
    varM = request.args.get("varM", None)    
    if varM is not None and gene is not None:
        return make_response(jsonify({"response": data_adaptor.data.var.set_index("name_0")[varM][gene]}), HTTPStatus.OK)
    else:
        return make_response(jsonify({"response": "NaN"}), HTTPStatus.OK)

def config_get(app_config, data_adaptor):
    config = get_client_config(app_config, data_adaptor)
    return make_response(jsonify(config), HTTPStatus.OK)


def userinfo_get(app_config, data_adaptor):
    config = get_client_userinfo(app_config, data_adaptor)
    return make_response(jsonify(config), HTTPStatus.OK)


def annotations_obs_get(request, data_adaptor):
    fields = request.args.getlist("annotation-name", None)
    num_columns_requested = len(data_adaptor.get_obs_keys()) if len(fields) == 0 else len(fields)
    if data_adaptor.server_config.exceeds_limit("column_request_max", num_columns_requested):
        return abort(HTTPStatus.BAD_REQUEST)
    preferred_mimetype = request.accept_mimetypes.best_match(["application/octet-stream"])
    if preferred_mimetype != "application/octet-stream":
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    try:
        labels = None
        annotations = data_adaptor.dataset_config.user_annotations
        if annotations.user_annotations_enabled():
            userID = _get_user_id(data_adaptor)
            labels = pickle.load(open(f"{userID}/obs.p","rb"))
        fbs = data_adaptor.annotation_to_fbs_matrix(Axis.OBS, fields, labels)
        return make_response(fbs, HTTPStatus.OK, {"Content-Type": "application/octet-stream"})
    except KeyError as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)


def annotations_put_fbs_helper(data_adaptor, fbs):
    """helper function to write annotations from fbs"""
    annotations = data_adaptor.dataset_config.user_annotations
    if not annotations.user_annotations_enabled():
        raise DisabledFeatureError("Writable annotations are not enabled")

    new_label_df = decode_matrix_fbs(fbs)
    if not new_label_df.empty:
        new_label_df = data_adaptor.check_new_labels(new_label_df)
    if not new_label_df.empty:
        userID = _get_user_id(data_adaptor)
        pickle.dump(new_label_df,open(f"{userID}/obs.p","wb"))



def inflate(data):
    return zlib.decompress(data)


def annotations_obs_put(request, data_adaptor):
    annotations = data_adaptor.dataset_config.user_annotations
    if not annotations.user_annotations_enabled():
        return abort(HTTPStatus.NOT_IMPLEMENTED)

    anno_collection = request.args.get("annotation-collection-name", default=None)
    fbs = inflate(request.get_data())

    if anno_collection is not None:
        if not annotations.is_safe_collection_name(anno_collection):
            return abort(HTTPStatus.BAD_REQUEST, "Bad annotation collection name")
        annotations.set_collection(anno_collection)

    try:
        annotations_put_fbs_helper(data_adaptor, fbs)
        res = json.dumps({"status": "OK"})
        return make_response(res, HTTPStatus.OK, {"Content-Type": "application/json"})
    except (ValueError, DisabledFeatureError, KeyError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)

def annotations_var_get(request, data_adaptor):
    fields = request.args.getlist("annotation-name", None)
    num_columns_requested = len(data_adaptor.get_var_keys()) if len(fields) == 0 else len(fields)
    if data_adaptor.server_config.exceeds_limit("column_request_max", num_columns_requested):
        return abort(HTTPStatus.BAD_REQUEST)
    preferred_mimetype = request.accept_mimetypes.best_match(["application/octet-stream"])
    if preferred_mimetype != "application/octet-stream":
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    try:
        labels = None
        return make_response(
            data_adaptor.annotation_to_fbs_matrix(Axis.VAR, fields, labels),
            HTTPStatus.OK,
            {"Content-Type": "application/octet-stream"},
        )
    except KeyError as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)


def data_var_put(request, data_adaptor):
    preferred_mimetype = request.accept_mimetypes.best_match(["application/octet-stream"])
    if preferred_mimetype != "application/octet-stream":
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    args = request.get_json()
    filter = args.get("filter",None)
    layer = args.get("layer","X")
    logscale = args.get("logscale","false")=="true"
    try:
        return make_response(
            data_adaptor.data_frame_to_fbs_matrix(filter, axis=Axis.VAR,layer=layer,logscale=logscale),
            HTTPStatus.OK,
            {"Content-Type": "application/octet-stream"},
        )
    except (FilterError, ValueError, ExceedsLimitError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)


def data_var_get(request, data_adaptor):
    preferred_mimetype = request.accept_mimetypes.best_match(["application/octet-stream"])
    if preferred_mimetype != "application/octet-stream":
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    try:
        layer = request.values.get("layer", default="X")
        logscale = request.values.get("logscale", default="false") == "true"
        args_filter_only = request.args.copy()
        args_filter_only.poplist("layer")  
        args_filter_only.poplist("logscale")        
        filter = _query_parameter_to_filter(args_filter_only)
        return make_response(
            data_adaptor.data_frame_to_fbs_matrix(filter, axis=Axis.VAR, layer=layer, logscale=logscale),
            HTTPStatus.OK,
            {"Content-Type": "application/octet-stream"},
        )
    except (FilterError, ValueError, ExceedsLimitError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)

def colors_get(data_adaptor):
    """
    if not data_adaptor.dataset_config.presentation__custom_colors:
        return make_response(jsonify({}), HTTPStatus.OK)
    try:
        return make_response(jsonify(data_adaptor.get_colors()), HTTPStatus.OK)
    except ColorFormatException as e:
        return abort_and_log(HTTPStatus.NOT_FOUND, str(e), include_exc_info=True)
    """
    return make_response(jsonify({}), HTTPStatus.OK)

def diffexp_obs_post(data, data_adaptor):
    if not data_adaptor.dataset_config.diffexp__enable:
        return abort(HTTPStatus.NOT_IMPLEMENTED)

    args = data.get_json()
    try:
        # TODO: implement varfilter mode
        mode = DiffExpMode(args["mode"])

        if mode == DiffExpMode.VAR_FILTER or "varFilter" in args:
            return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, "varFilter not enabled")

        set1_filter = args.get("set1", {"filter": {}})["filter"]
        set2_filter = args.get("set2", {"filter": {}})["filter"]
        count = args.get("count", None)

        if set1_filter is None or set2_filter is None or count is None:
            return abort_and_log(HTTPStatus.BAD_REQUEST, "missing required parameter")
        if Axis.VAR in set1_filter or Axis.VAR in set2_filter:
            return abort_and_log(HTTPStatus.BAD_REQUEST, "var axis filter not enabled")

    except (KeyError, TypeError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)

    try:
        diffexp = data_adaptor.diffexp_topN(set1_filter, set2_filter, count)
        return make_response(diffexp, HTTPStatus.OK, {"Content-Type": "application/json"})
    except (ValueError, DisabledFeatureError, FilterError, ExceedsLimitError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)
    except JSONEncodingValueError:
        # JSON encoding failure, usually due to bad data. Just let it ripple up
        # to default exception handler.
        current_app.logger.warning(JSON_NaN_to_num_warning_msg)
        raise  

def layout_obs_get(request, data_adaptor):
    fields = request.args.getlist("layout-name", None)
    num_columns_requested = len(data_adaptor.get_embedding_names()) if len(fields) == 0 else len(fields)
    if data_adaptor.server_config.exceeds_limit("column_request_max", num_columns_requested):
        return abort(HTTPStatus.BAD_REQUEST)

    preferred_mimetype = request.accept_mimetypes.best_match(["application/octet-stream"])
    if preferred_mimetype != "application/octet-stream":
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    try:
        return make_response(
            data_adaptor.layout_to_fbs_matrix(fields), HTTPStatus.OK, {"Content-Type": "application/octet-stream"}
        )
    except (KeyError, DatasetAccessError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)
    except PrepareError:
        return abort_and_log(
            HTTPStatus.NOT_IMPLEMENTED,
            f"No embedding available {request.path}",
            loglevel=logging.ERROR,
            include_exc_info=True,
        )

def sankey_data_put(request, data_adaptor):
    args = request.get_json()
    labels = args.get("labels", None)
    name = args.get("name", None)
    filter = args.get("filter",None)

    if not labels:
        return make_response(jsonify({"edges":[],"weights":[]}),HTTPStatus.OK, {"Content-Type": "application/json"})
    

    try:
        userID = _get_user_id(data_adaptor)          
        edges,weights = data_adaptor.compute_sankey_df(labels,name,filter, userID)
        edges = [list(x) for x in edges]
        weights = list(weights)
        return make_response(jsonify({"edges": edges, "weights": weights}), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)

def save_metadata_put(request, data_adaptor):
    args = request.get_json()
    labelNames = args.get("labelNames",None)
    filter = args["filter"] if args else None
    
    try:
        shape = data_adaptor.get_shape()
        obs_mask = data_adaptor._axis_filter_to_mask(Axis.OBS, filter["obs"], shape[0])
    except (KeyError, IndexError):
        raise FilterError("Error parsing filter")

    userID = _get_user_id(data_adaptor)        


    labels = pickle.load(open(f"{userID}/obs.p","rb"))
    labels = labels[labelNames]
    labels = labels[obs_mask]
    labels.to_csv(f"{userID}/{userID}_obs.csv")

    @after_this_request
    def remove_file(response):
        try:
            os.remove(f"{userID}/{userID}_obs.csv")
        except Exception as error:
            print(error)
        return response

    try:
        direc = pathlib.Path().absolute()
        return send_file(f"{direc}/{userID}/{userID}_obs.csv",as_attachment=True)
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)  

def save_genedata_put(request, data_adaptor):
    userID = _get_user_id(data_adaptor)        
    try:
        direc = pathlib.Path().absolute()
        return send_file(f"{direc}/{userID}/gene-sets.csv",as_attachment=True)
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)  


def _get_user_id(data_adaptor):
    annotations = data_adaptor.dataset_config.user_annotations        
    userID = f"{annotations._get_userdata_idhash(data_adaptor)}"       
    return userID

def delete_obsm_put(request, data_adaptor):
    args = request.get_json()
    embNames = args.get("embNames",None)
    fail=False
    userID = _get_user_id(data_adaptor)
    if embNames is not None:
        for embName in embNames:
            
            if os.path.exists(f"{userID}/emb/{embName}.p"):
                os.remove(f"{userID}/emb/{embName}.p")
            if os.path.exists(f"{userID}/nnm/{embName}.p"):
                os.remove(f"{userID}/nnm/{embName}.p")
            if os.path.exists(f"{userID}/params/{embName}.p"):
                os.remove(f"{userID}/params/{embName}.p")                
    try:
        return make_response(jsonify({"fail": fail}), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)    

def initialize_user(data_adaptor):
    userID = _get_user_id(data_adaptor)
    data_adaptor._initialize_user_folders(userID)           

    try:
        return make_response(jsonify({"fail": False}), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)    


def rename_obsm_put(request, data_adaptor):
    args = request.get_json()
    embNames = args.get("embNames",None)
    oldName = args.get("oldName",None)
    newName = args.get("newName",None)
    userID = _get_user_id(data_adaptor)
    if embNames is not None and oldName is not None and newName is not None:
        for embName in embNames:
            os.rename(f"{userID}/emb/{embName}.p",f"{userID}/emb/{embName.replace(oldName,newName)}.p")
            os.rename(f"{userID}/nnm/{embName}.p",f"{userID}/nnm/{embName.replace(oldName,newName)}.p")
            os.rename(f"{userID}/params/{embName}.p",f"{userID}/params/{embName.replace(oldName,newName)}.p")
    try:
        return make_response(jsonify({"schema": schema_get_helper(data_adaptor)}), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True) 

def leiden_put(request, data_adaptor):
    args = request.get_json()
    name = args.get("name", None)
    cName = args.get("cName", None)
    resolution = args.get('resolution',1.0)
    filter = args.get('filter',None)
    try:
        userID = _get_user_id(data_adaptor)
        cl = list(data_adaptor.compute_leiden(name,cName,resolution,filter,userID))
        return make_response(jsonify({"clusters": cl}), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)


def layout_obs_put(request, data_adaptor):
    args = request.get_json()
    filter = args["filter"] if args else None
    if not filter:
        return abort_and_log(HTTPStatus.BAD_REQUEST, "obs filter is required")
    method = args["method"] if args else "umap"
    reembedParams = args["params"] if args else {}
    parentName = args["parentName"] if args else ""
    embName = args["embName"] if args else None

    try:
        userID = _get_user_id(data_adaptor)                 
        schema = data_adaptor.compute_embedding(method, filter, reembedParams, parentName, embName, userID, hosted = current_app.hosted_mode)
        return make_response(jsonify(schema), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)


def preprocess_put(request, data_adaptor):
    args = request.get_json()
    reembedParams = args["params"] if args else {}
    filter = args["filter"] if args else None
    if not filter:
        return abort_and_log(HTTPStatus.BAD_REQUEST, "obs filter is required")
        
    try:
        userID = _get_user_id(data_adaptor)            
        data_adaptor.compute_preprocess(reembedParams, filter, userID)
        return make_response(jsonify(schema_get_helper(data_adaptor)), HTTPStatus.OK, {"Content-Type": "application/json"})
    except NotImplementedError as e:
        return abort_and_log(HTTPStatus.NOT_IMPLEMENTED, str(e))
    except (ValueError, DisabledFeatureError, FilterError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)


def reembed_parameters_get(request, data_adaptor):
    preferred_mimetype = request.accept_mimetypes.best_match(["application/json", "text/csv"])
    if preferred_mimetype not in ("application/json", "text/csv"):
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    try:
        annotations = data_adaptor.dataset_config.user_annotations
        (reembedParams) = annotations.read_reembed_parameters(data_adaptor)
        return make_response(
            jsonify({"reembedParams": reembedParams}), HTTPStatus.OK
        )
    except (ValueError, KeyError, AnnotationsError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e))

def reembed_parameters_obsm_put(request, data_adaptor):
    embName = request.get_json()["embName"]
    userID = _get_user_id(data_adaptor)
    try:
        reembedParams = pickle.load(open(f"{userID}/params/{embName}","rb"))
    except:
        reembedParams = None

    if reembedParams is not None:
        reembedParams = reembedParams.copy()
        try:
            del reembedParams['parentParams']
        except:
            pass
        try:
            del reembedParams['sample_ids']
        except:
            pass
        try:
            del reembedParams['feature_ids']
        except:
            pass
        try:
            del reembedParams['feature_weights']
        except:
            pass

    try:
        if reembedParams is not None:
            for k in reembedParams.keys():
                if type(reembedParams[k]).__module__ == 'numpy':
                    reembedParams[k] = reembedParams[k].item()
            return make_response(
                jsonify({"reembedParams": reembedParams}), HTTPStatus.OK
            )
        else:
            return make_response(
                jsonify({"reembedParams": {}}), HTTPStatus.OK
            )            
    except (ValueError, KeyError, AnnotationsError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e))

def genesets_get(request, data_adaptor):
    preferred_mimetype = request.accept_mimetypes.best_match(["application/json", "text/csv"])
    if preferred_mimetype not in ("application/json", "text/csv"):
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    try:
        annotations = data_adaptor.dataset_config.user_annotations
        (genesets, tid) = annotations.read_gene_sets(data_adaptor)

        if preferred_mimetype == "text/csv":
            return make_response(
                annotations.gene_sets_to_csv(genesets),
                HTTPStatus.OK,
                {
                    "Content-Type": "text/csv",
                    "Content-Disposition": "attachment; filename=genesets.csv",
                },
            )
        else:
            return make_response(
                jsonify({"genesets": annotations.gene_sets_to_response(genesets), "tid": tid}), HTTPStatus.OK
            )
    except (ValueError, KeyError, AnnotationsError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e))

def reembed_parameters_put(request, data_adaptor):
    annotations = data_adaptor.dataset_config.user_annotations

    anno_collection = request.args.get("annotation-collection-name", default=None)
    if anno_collection is not None:
        if not annotations.is_safe_collection_name(anno_collection):
            return abort(HTTPStatus.BAD_REQUEST, "Bad annotation collection name")
        annotations.set_collection(anno_collection)

    args = request.get_json()
    try:
        reembedParams = args.get("reembedParams", None)
        if reembedParams is None:
            abort(HTTPStatus.BAD_REQUEST)

        annotations.write_reembed_parameters(reembedParams, data_adaptor)
        return make_response(jsonify({"status": "OK"}), HTTPStatus.OK)
    except (ValueError, DisabledFeatureError, KeyError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)
    except (ObsoleteRequest, TypeError) as e:
        return abort(HTTPStatus.NOT_FOUND, description=str(e))


def genesets_put(request, data_adaptor):
    annotations = data_adaptor.dataset_config.user_annotations
    if not annotations.gene_sets_save_enabled():
        return abort(HTTPStatus.NOT_IMPLEMENTED)

    anno_collection = request.args.get("annotation-collection-name", default=None)
    if anno_collection is not None:
        if not annotations.is_safe_collection_name(anno_collection):
            return abort(HTTPStatus.BAD_REQUEST, "Bad annotation collection name")
        annotations.set_collection(anno_collection)

    args = request.get_json()
    try:
        genesets = args.get("genesets", None)
        tid = args.get("tid", None)
        if genesets is None:
            abort(HTTPStatus.BAD_REQUEST)

        annotations.write_gene_sets(genesets, tid, data_adaptor)
        return make_response(jsonify({"status": "OK"}), HTTPStatus.OK)
    except (ValueError, DisabledFeatureError, KeyError) as e:
        return abort_and_log(HTTPStatus.BAD_REQUEST, str(e), include_exc_info=True)
    except (ObsoleteRequest, TypeError) as e:
        return abort(HTTPStatus.NOT_FOUND, description=str(e))


def summarize_var_helper(request, data_adaptor, key, raw_query):
    preferred_mimetype = request.accept_mimetypes.best_match(["application/octet-stream"])
    if preferred_mimetype != "application/octet-stream":
        return abort(HTTPStatus.NOT_ACCEPTABLE)

    summary_method = request.values.get("method", default="mean")
    query_hash = summarizeQueryHash(raw_query)
    if key and query_hash != key:
        return abort(HTTPStatus.BAD_REQUEST, description="query key did not match")

    args_filter_only = request.values.copy()
    args_filter_only.poplist("method")
    args_filter_only.poplist("key")
    args_filter_only.poplist("layer")
    args_filter_only.poplist("logscale")

    try:
        layer = request.values.get("layer", default="X")  
        logscale = request.values.get("logscale", default="false")=="true"        
        filter = _query_parameter_to_filter(args_filter_only)
        return make_response(
            data_adaptor.summarize_var(summary_method, filter, query_hash, layer, logscale),
            HTTPStatus.OK,
            {"Content-Type": "application/octet-stream"},
        )
    except (ValueError) as e:
        return abort(HTTPStatus.NOT_FOUND, description=str(e))
    except (UnsupportedSummaryMethod, FilterError) as e:
        return abort(HTTPStatus.BAD_REQUEST, description=str(e))


def summarize_var_get(request, data_adaptor):
    return summarize_var_helper(request, data_adaptor, None, request.query_string)


def summarize_var_post(request, data_adaptor):
    if not request.content_type or "application/x-www-form-urlencoded" not in request.content_type:
        return abort(HTTPStatus.UNSUPPORTED_MEDIA_TYPE)
    if request.content_length > 1_000_000:  # just a sanity check to avoid memory exhaustion
        return abort(HTTPStatus.BAD_REQUEST)

    key = request.args.get("key", default=None)
    return summarize_var_helper(request, data_adaptor, key, request.get_data())
